@@ -1,7 +1,7 @@
 /* asmstub.c - a version of shared_src/asm.S that works under Unix */
 /*
  *  GRUB  --  GRand Unified Bootloader
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2000  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,13 +26,9 @@
 /* Simulator entry point. */
 int grub_stage2 (void);
 
-/* We want to prevent any circularararity in our stubs, as well as
-   libc name clashes. */
-#define WITHOUT_LIBC_STUBS 1
-#include "shared.h"
-
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <assert.h>
 #include <stdio.h>
 #include <sys/types.h>
@@ -43,45 +39,42 @@ int grub_stage2 (void);
 #include <string.h>
 #include <unistd.h>
 #include <setjmp.h>
+#include <sys/time.h>
+#include <termios.h>
 
 #ifdef __linux__
 # include <sys/ioctl.h>		/* ioctl */
-# include <linux/hdreg.h>	/* HDIO_GETGEO */
-# if __GLIBC__ < 2
+# if !defined(__GLIBC__) || \
+	((__GLIBC__ < 2) || ((__GLIBC__ == 2) && (__GLIBC_MINOR__ < 1)))
 /* Maybe libc doesn't have large file support.  */
 #  include <linux/unistd.h>	/* _llseek */
-#  include <linux/fs.h>		/* BLKFLSBUF */
-# endif /* GLIBC < 2 */
+# endif /* (GLIBC < 2) || ((__GLIBC__ == 2) && (__GLIBC_MINOR < 1)) */
 # ifndef BLKFLSBUF
-#  define BLKFLSBUF	_IO (0x12,97)
+#  define BLKFLSBUF	_IO (0x12,97)	/* flush buffer cache */
 # endif /* ! BLKFLSBUF */
 #endif /* __linux__ */
 
-#ifdef __FreeBSD__
-# include <sys/disklabel.h>
-#endif /* __FreeBSD__ */
+/* We want to prevent any circularararity in our stubs, as well as
+   libc name clashes. */
+#define WITHOUT_LIBC_STUBS 1
+#include <shared.h>
+#include <device.h>
+#include <serial.h>
 
 /* Simulated memory sizes. */
-#define EXTENDED_MEMSIZE (4 * 1024 * 1024)	/* 4MB */
-#define CONVENTIONAL_MEMSIZE (640) /* 640kB */
-
-/* Simulated disk sizes. */
-#define DEFAULT_FD_CYLINDERS	80
-#define DEFAULT_FD_HEADS	2
-#define DEFAULT_FD_SECTORS	18
-#define DEFAULT_HD_CYLINDERS	620
-#define DEFAULT_HD_HEADS	128
-#define DEFAULT_HD_SECTORS	63
+#define EXTENDED_MEMSIZE (3 * 1024 * 1024)	/* 3MB */
+#define CONVENTIONAL_MEMSIZE (640 * 1024)	/* 640kB */
 
 unsigned long install_partition = 0x20000;
 unsigned long boot_drive = 0;
+int saved_entryno = 0;
 char version_string[] = VERSION;
 char config_file[128] = "/boot/grub/menu.lst"; /* FIXME: arbitrary */
+unsigned long linux_text_len = 0;
 
 /* Emulation requirements. */
 char *grub_scratch_mem = 0;
 
-#define NUM_DISKS 256
 struct geometry *disks = 0;
 
 /* The map between BIOS drives and UNIX device file names.  */
@@ -90,10 +83,16 @@ char **device_map = 0;
 /* The jump buffer for exiting correctly.  */
 static jmp_buf env_for_exit;
 
-/* Forward declarations.  */
-static void get_floppy_disk_name (char *name, int unit);
-static void get_ide_disk_name (char *name, int unit);
-static void get_scsi_disk_name (char *name, int unit);
+/* The file descriptor for a serial device.  */
+static int serial_fd = -1;
+
+/* The file name of a serial device.  */
+static char *serial_device = 0;
+
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+/* The speed of a serial device.  */
+static unsigned int serial_speed;
+#endif /* SIMULATE_SLOWNESS_OF_SERIAL */
 
 /* The main entry point into this mess. */
 int
@@ -102,8 +101,8 @@ grub_stage2 (void)
   /* These need to be static, because they survive our stack transitions. */
   static int status = 0;
   static char *realstack;
-  int i, num_hd = 0;
   char *scratch, *simstack;
+  int i;
 
   /* We need a nested function so that we get a clean stack frame,
      regardless of how the code is optimized. */
@@ -145,60 +144,9 @@ grub_stage2 (void)
   for (i = 0; i < NUM_DISKS; i++)
     disks[i].flags = -1;
 
-  assert (device_map == 0);
-  device_map = malloc (NUM_DISKS * sizeof (char *));
-  assert (device_map);
-
-  /* Probe devices for creating the device map.  */
-
-  /* Initialize DEVICE_MAP.  */
-  for (i = 0; i < NUM_DISKS; i++)
-    device_map[i] = 0;
-
-  /* Print something as the user does not think GRUB has been crashed.  */
-  fprintf (stderr,
-	   "Probe devices to guess BIOS drives. This may take a long time.\n");
-    
-  /* Floppies.  */
-  if (! no_floppy)
-    for (i = 0; i < 2; i++)
-      {
-	char name[16];
-
-	if (i == 1 && ! probe_second_floppy)
-	  break;
-	
-	get_floppy_disk_name (name, i);
-	if (check_device (name))
-	  assign_device_name (i, name);
-      }
+  if (! init_device_map (&device_map, device_map_file, floppy_disks))
+    return 1;
   
-  /* IDE disks.  */
-  for (i = 0; i < 4; i++)
-    {
-      char name[16];
-
-      get_ide_disk_name (name, i);
-      if (check_device (name))
-	{
-	  assign_device_name (num_hd + 0x80, name);
-	  num_hd++;
-	}
-    }
-  
-  /* The rest is SCSI disks.  */
-  for (i = 0; i < 8; i++)
-    {
-      char name[16];
-
-      get_scsi_disk_name (name, i);
-      if (check_device (name))
-	{
-	  assign_device_name (num_hd + 0x80, name);
-	  num_hd++;
-	}
-    }
-
   /* Check some invariants. */
   assert ((SCRATCHSEG << 4) == SCRATCHADDR);
   assert ((BUFFERSEG << 4) == BUFFERADDR);
@@ -222,14 +170,14 @@ grub_stage2 (void)
 
   /* Make sure that actual writing is done.  */
   sync ();
-  
+
   /* Set our stack, and go for it. */
   simstack = (char *) PROTSTACKINIT;
   doit ();
 
   /* I don't know if this is necessary really.  */
   sync ();
-  
+
 #ifdef HAVE_LIBCURSES
   if (use_curses)
     endwin ();
@@ -249,119 +197,25 @@ grub_stage2 (void)
 	close (disks[i].flags);
       }
 
+  if (serial_fd >= 0)
+    close (serial_fd);
+  
   /* Release memory. */
-  for (i = 0; i < NUM_DISKS; i++)
-    free (device_map[i]);
-  free (device_map[i]);
+  restore_device_map (device_map);
   device_map = 0;
   free (disks);
   disks = 0;
   free (scratch);
   grub_scratch_mem = 0;
 
+  if (serial_device)
+    free (serial_device);
+  serial_device = 0;
+  
   /* Ahh... at last we're ready to return to caller. */
   return status;
 }
 
-/* These three functions are quite different among OSes.  */
-static void
-get_floppy_disk_name (char *name, int unit)
-{
-#if defined(__linux__) || defined(__GNU__)
-  /* GNU/Linux and GNU/Hurd */
-  sprintf (name, "/dev/fd%d", unit);
-#elif defined(__FreeBSD__)
-  /* FreeBSD */
-  sprintf (name, "/dev/rfd%d", unit);
-#else
-# warning "BIOS drives cannot be guessed in your operating system."
-  /* Set NAME to a bogus string.  */
-  *name = 0;
-#endif
-}
-
-static void
-get_ide_disk_name (char *name, int unit)
-{
-#if defined(__linux__)
-  /* GNU/Linux */
-  sprintf (name, "/dev/hd%c", unit + 'a');
-#elif defined(__GNU__)
-  /* GNU/Hurd */
-  sprintf (name, "/dev/hd%d", unit);
-#elif defined(__FreeBSD__)
-  /* FreeBSD */
-  sprintf (name, "/dev/rwd%d", unit);
-#else
-# warning "BIOS drives cannot be guessed in your operating system."
-  /* Set NAME to a bogus string.  */
-  *name = 0;
-#endif
-}
-
-static void
-get_scsi_disk_name (char *name, int unit)
-{
-#if defined(__linux__)
-  /* GNU/Linux */
-  sprintf (name, "/dev/sd%c", unit + 'a');
-#elif defined(__GNU__)
-  /* GNU/Hurd */
-  sprintf (name, "/dev/sd%d", unit);
-#elif defined(__FreeBSD__)
-  /* FreeBSD */
-  sprintf (name, "/dev/rda%d", unit);
-#else
-# warning "BIOS drives cannot be guessed in your operating system."
-  /* Set NAME to a bogus string.  */
-  *name = 0;
-#endif
-}
-  
-/* Check if DEVICE can be read. If an error occurs, return zero,
-   otherwise return non-zero.  */
-int
-check_device (const char *device)
-{
-  char buf[512];
-  FILE *fp;
-
-  fp = fopen (device, "r");
-  if (! fp)
-    {
-      switch (errno)
-	{
-#ifdef ENOMEDIUM
-	case ENOMEDIUM:
-# if 0
-	  /* At the moment, this finds only CDROMs, which can't be
-	     read anyway, so leave it out. Code should be
-	     reactivated if `removable disks' and CDROMs are
-	     supported.  */
-	  /* Accept it, it may be inserted.  */
-	  return 1;
-# endif
-	  break;
-#endif /* ENOMEDIUM */
-	default:
-	  /* Break case and leave.  */
-	  break;
-	}
-      /* Error opening the device.  */
-      return 0;
-    }
-  
-  /* Attempt to read the first sector.  */
-  if (fread (buf, 1, 512, fp) != 512)
-    {
-      fclose (fp);
-      return 0;
-    }
-  
-  fclose (fp);
-  return 1;
-}
-
 /* Assign DRIVE to a device name DEVICE.  */
 void
 assign_device_name (int drive, const char *device)
@@ -369,16 +223,19 @@ assign_device_name (int drive, const char *device)
   /* If DRIVE is already assigned, free it.  */
   if (device_map[drive])
     free (device_map[drive]);
-  
+
   /* If the old one is already opened, close it.  */
   if (disks[drive].flags != -1)
     {
       close (disks[drive].flags);
       disks[drive].flags = -1;
     }
-  
+
   /* Assign DRIVE to DEVICE.  */
-  device_map[drive] = strdup (device);
+  if (! device)
+    device_map[drive] = 0;
+  else
+    device_map[drive] = strdup (device);
 }
 
 void
@@ -393,17 +250,29 @@ stop (void)
   longjmp (env_for_exit, 1);
 }
 
+void
+grub_reboot (void)
+{
+  stop ();
+}
+
+void
+grub_halt (int no_apm)
+{
+  stop ();
+}
 
 /* calls for direct boot-loader chaining */
 void
-chain_stage1 (int segment, int offset, int part_table_addr)
+chain_stage1 (unsigned long segment, unsigned long offset,
+	      unsigned long part_table_addr)
 {
   stop ();
 }
 
 
 void
-chain_stage2 (int segment, int offset)
+chain_stage2 (unsigned long segment, unsigned long offset)
 {
   stop ();
 }
@@ -427,7 +296,7 @@ big_linux_boot (void)
 
 /* booting a multiboot executable */
 void
-multi_boot (int start, int mbi)
+multi_boot (int start, int mb_info)
 {
   stop ();
 }
@@ -439,6 +308,30 @@ gateA20 (int linear)
   /* Nothing to do in the simulator. */
 }
 
+/* Set up the int15 handler.  */
+void
+set_int15_handler (void)
+{
+  /* Nothing to do in the simulator.  */
+}
+
+/* Restore the original int15 handler.  */
+void
+unset_int15_handler (void)
+{
+  /* Nothing to do in the simulator.  */
+}
+
+/* The key map.  */
+unsigned short bios_key_map[KEY_MAP_SIZE + 1];
+unsigned short ascii_key_map[KEY_MAP_SIZE + 1];
+
+/* Copy MAP to the drive map and set up the int13 handler.  */
+void
+set_int13_handler (unsigned short *map)
+{
+  /* Nothing to do in the simulator.  */
+}
 
 int
 get_code_end (void)
@@ -452,10 +345,10 @@ get_code_end (void)
 int
 get_memsize (int type)
 {
-  if (!type)
-    return CONVENTIONAL_MEMSIZE;
+  if (! type)
+    return CONVENTIONAL_MEMSIZE >> 10;
   else
-    return EXTENDED_MEMSIZE;
+    return EXTENDED_MEMSIZE >> 10;
 }
 
 
@@ -466,7 +359,7 @@ get_memsize (int type)
 int
 get_eisamemsize (void)
 {
-  return (EXTENDED_MEMSIZE / 1024);
+  return (EXTENDED_MEMSIZE >> 10);
 }
 
 
@@ -475,26 +368,58 @@ get_eisamemsize (void)
 #define MMAR_DESC_TYPE_ACPI_RECLAIM 3 /* usable by OS after reading ACPI */
 #define MMAR_DESC_TYPE_ACPI_NVS 4 /* required to save between NVS sessions */
 
+#define MMAR_DESC_LENGTH	20
+
 /* Fetch the next entry in the memory map and return the continuation
    value.  DESC is a pointer to the descriptor buffer, and CONT is the
    previous continuation value (0 to get the first entry in the
-   map). */
+   map).  */
 int
 get_mmap_entry (struct mmar_desc *desc, int cont)
 {
-  if (! cont)
+  /* Record the memory map statically.  */
+  static struct mmar_desc desc_table[] =
+  {
+    /* The conventional memory.  */
     {
-      /* First entry, located at 1MB. */
-      desc->desc_len = sizeof (*desc) - sizeof (desc->desc_len);
-      desc->addr = 1024 * 1024;
-      desc->length = EXTENDED_MEMSIZE;
-      desc->type = MMAR_DESC_TYPE_AVAILABLE;
+      MMAR_DESC_LENGTH,
+      0,
+      CONVENTIONAL_MEMSIZE,
+      MMAR_DESC_TYPE_AVAILABLE
+    },
+    /* BIOS RAM and ROM (such as video memory).  */
+    {
+      MMAR_DESC_LENGTH,
+      CONVENTIONAL_MEMSIZE,
+      0x100000 - CONVENTIONAL_MEMSIZE,
+      MMAR_DESC_TYPE_RESERVED
+    },
+    /* The extended memory.  */
+    {
+      MMAR_DESC_LENGTH,
+      0x100000,
+      EXTENDED_MEMSIZE,
+      MMAR_DESC_TYPE_AVAILABLE
+    }
+  };
+  
+  int num = sizeof (desc_table) / sizeof (*desc_table);
+
+  if (cont < 0 || cont >= num)
+    {
+      /* Should not happen.  */
+      desc->desc_len = 0;
     }
   else
     {
-      /* No more entries, so give an error. */
-      desc->desc_len = 0;
+      /* Copy the entry.  */
+      *desc = desc_table[cont++];
+
+      /* If the next entry exists, return the index.  */
+      if (cont < num)
+	return cont;
     }
+  
   return 0;
 }
 
@@ -507,10 +432,35 @@ getrtsecs (void)
   return time (0);
 }
 
+int
+currticks (void)
+{
+  struct timeval tv;
+  long csecs;
+  int ticks_per_csec, ticks_per_usec;
+
+  /* Note: 18.2 ticks/sec.  */
+
+  /* Get current time.  */
+  gettimeofday (&tv, 0);
+
+  /* Compute centiseconds.  */
+  csecs = tv.tv_sec / 10;
+
+  /* Ticks per centisecond.  */
+  ticks_per_csec = csecs * 182;
+
+  /* Ticks per microsecond.  */
+  ticks_per_usec = (((tv.tv_sec - csecs * 10) * 1000000 + tv.tv_usec)
+		    * 182 / 10000000);
+
+  /* Sum them.  */
+  return ticks_per_csec + ticks_per_usec;
+}
 
 /* low-level character I/O */
 void
-cls (void)
+console_cls (void)
 {
 #ifdef HAVE_LIBCURSES
   if (use_curses)
@@ -521,7 +471,7 @@ cls (void)
 
 /* returns packed values, LSB+1 is x, LSB is y */
 int
-getxy (void)
+console_getxy (void)
 {
   int y, x;
 #ifdef HAVE_LIBCURSES
@@ -535,7 +485,7 @@ getxy (void)
 
 
 void
-gotoxy (int x, int y)
+console_gotoxy (int x, int y)
 {
 #ifdef HAVE_LIBCURSES
   if (use_curses)
@@ -543,15 +493,34 @@ gotoxy (int x, int y)
 #endif
 }
 
-
 /* displays an ASCII character.  IBM displays will translate some
    characters to special graphical ones */
 void
-grub_putchar (int c)
+console_putchar (int c)
 {
 #ifdef HAVE_LIBCURSES
   if (use_curses)
-    addch (c);
+    {
+      /* In ncurses, a newline is treated badly, so we emulate it in our
+	 own way.  */
+      if (c == '\n')
+	{
+	  int x, y;
+
+	  getyx (stdscr, y, x);
+	  if (y + 1 == LINES)
+	    scroll (stdscr);
+	  else
+	    move (y + 1, x);
+	}
+      else
+	{
+	  addch (c);
+#ifdef REFRESH_IMMEDIATELY
+	  refresh ();
+#endif
+	}
+    }
   else
 #endif
   putchar (c);
@@ -567,13 +536,13 @@ static int save_char = ERR;
 
 /* returns packed BIOS/ASCII code */
 int
-getkey (void)
+console_getkey (void)
 {
+  int c;
+
 #ifdef HAVE_LIBCURSES
   if (use_curses)
     {
-      int c;
-
       /* If checkkey has already got a character, then return it.  */
       if (save_char != ERR)
 	{
@@ -581,21 +550,25 @@ getkey (void)
 	  save_char = ERR;
 	  return c;
 	}
-      
+
       wtimeout (stdscr, -1);
       c = getch ();
       wtimeout (stdscr, 100);
-      return c;
     }
+  else
 #endif
+    c = getchar ();
 
-  return getchar ();
+  /* Quit if we get EOF. */
+  if (c == -1)
+    stop ();
+  return c;
 }
 
 
 /* like 'getkey', but doesn't wait, returns -1 if nothing available */
 int
-checkkey (void)
+console_checkkey (void)
 {
 #ifdef HAVE_LIBCURSES
   if (use_curses)
@@ -606,7 +579,7 @@ checkkey (void)
 	 means checkkey is called twice continuously.  */
       if (save_char != ERR)
 	return save_char;
-      
+
       c = getch ();
       /* If C is not ERR, then put it back in the input queue.  */
       if (c != ERR)
@@ -615,7 +588,8 @@ checkkey (void)
     }
 #endif
 
-  /* Just pretend they hit the space bar. */
+  /* Just pretend they hit the space bar, then read the real key when
+     they call getkey. */
   return ' ';
 }
 
@@ -638,48 +612,6 @@ set_attrib (int attr)
 #endif
 }
 
-
-/* Get the geometry of a drive DRIVE. If an error occurs, return zero,
-   otherwise non-zero.  */
-static int
-get_drive_geometry (int drive)
-{
-  struct geometry *geom = &disks[drive];
-
-#if defined(__linux__)
-  /* Linux */
-  struct hd_geometry hdg;
-  if (ioctl (geom->flags, HDIO_GETGEO, &hdg))
-    return 0;
-  
-  /* Got the geometry, so save it. */
-  geom->cylinders = hdg.cylinders;
-  geom->heads = hdg.heads;
-  geom->sectors = hdg.sectors;
-  /* FIXME: Should get the real number of sectors.  */
-  geom->total_sectors = hdg.cylinders * hdg.heads * hdg.sectors;
-  return 1;
-  
-#elif defined(__FreeBSD__)
-  /* FreeBSD */
-  struct disklabel hdg;
-  if (ioctl (disks[drive].flags, DIOCGDINFO, &hdg))
-    return 0;
-  
-  disks[drive].cylinders = hdg.d_ncylinders;
-  disks[drive].heads = hdg.d_ntracks;
-  disks[drive].sectors = hdg.d_nsectors;
-  disks[drive].total_sectors = hdg.d_secperunit;
-  return 1;
-  
-#else
-# warning "In your operating system, automatic detection of geometries \
-will not be performed."
-  return 0;
-#endif
-}
-
-  
 /* Low-level disk I/O.  Our stubbed version just returns a file
    descriptor, not the actual geometry. */
 int
@@ -703,21 +635,27 @@ get_diskinfo (int drive, struct geometry *geometry)
       if (verbose)
 	grub_printf ("Attempt to open drive 0x%x (%s)\n",
 		     drive, devname);
-	  
+
       /* Open read/write, or read-only if that failed. */
       if (! read_only)
 	disks[drive].flags = open (devname, O_RDWR);
-      
+
       if (disks[drive].flags == -1)
 	{
 	  if (read_only || errno == EACCES || errno == EROFS)
 	    {
 	      disks[drive].flags = open (devname, O_RDONLY);
 	      if (disks[drive].flags == -1)
-		return -1;
+		{
+		  assign_device_name (drive, 0);
+		  return -1;
+		}
 	    }
 	  else
-	    return -1;
+	    {
+	      assign_device_name (drive, 0);
+	      return -1;
+	    }
 	}
 
       /* Attempt to read the first sector.  */
@@ -725,41 +663,23 @@ get_diskinfo (int drive, struct geometry *geometry)
 	{
 	  close (disks[drive].flags);
 	  disks[drive].flags = -1;
+	  assign_device_name (drive, 0);
 	  return -1;
 	}
 
       if (disks[drive].flags != -1)
-	{
-	  if (! get_drive_geometry (drive))
-	    {
-	      /* Set some arbitrary defaults. */
-	      if (drive & 0x80)
-		{
-		  /* Hard drive. */
-		  disks[drive].cylinders = DEFAULT_HD_CYLINDERS;
-		  disks[drive].heads = DEFAULT_HD_HEADS;
-		  disks[drive].sectors = DEFAULT_HD_SECTORS;
-		  disks[drive].total_sectors = (DEFAULT_HD_CYLINDERS
-						* DEFAULT_HD_HEADS
-						* DEFAULT_HD_SECTORS);
-		}
-	      else
-		{
-		  /* Floppy. */
-		  disks[drive].cylinders = DEFAULT_FD_CYLINDERS;
-		  disks[drive].heads = DEFAULT_FD_HEADS;
-		  disks[drive].sectors = DEFAULT_FD_SECTORS;
-		  disks[drive].total_sectors = (DEFAULT_FD_CYLINDERS
-						* DEFAULT_FD_HEADS
-						* DEFAULT_FD_SECTORS);
-		}
-	    }
-	}
+	get_drive_geometry (&disks[drive], device_map, drive);
     }
 
   if (disks[drive].flags == -1)
     return -1;
 
+#ifdef __linux__
+  /* In Linux, invalidate the buffer cache, so that left overs
+     from other program in the cache are flushed and seen by us */
+  ioctl (disks[drive].flags, BLKFLSBUF, 0);
+#endif
+
   *geometry = disks[drive];
   return 0;
 }
@@ -770,7 +690,7 @@ static int
 nread (int fd, char *buf, size_t len)
 {
   int size = len;
-  
+
   while (len)
     {
       int ret = read (fd, buf, len);
@@ -786,7 +706,7 @@ nread (int fd, char *buf, size_t len)
       len -= ret;
       buf += ret;
     }
-  
+
   return size;
 }
 
@@ -796,7 +716,7 @@ static int
 nwrite (int fd, char *buf, size_t len)
 {
   int size = len;
-  
+
   while (len)
     {
       int ret = write (fd, buf, len);
@@ -812,7 +732,7 @@ nwrite (int fd, char *buf, size_t len)
       len -= ret;
       buf += ret;
     }
-  
+
   return size;
 }
 
@@ -829,7 +749,7 @@ hex_dump (void *buf, size_t size)
   int column = 0;
   /* how many bytes written */
   int count = 0;
-  
+
   while (size > 0)
     {
       /* high 4 bits */
@@ -877,12 +797,14 @@ biosdisk (int subfunc, int drive, struct geometry *geometry,
     return BIOSDISK_ERROR_GEOMETRY;
 
   /* Seek to the specified location. */
-#if defined(__linux) && (__GLIBC__ < 2)
+#if defined(__linux__) && (!defined(__GLIBC__) || \
+	((__GLIBC__ < 2) || ((__GLIBC__ == 2) && (__GLIBC_MINOR__ < 1))))
   /* Maybe libc doesn't have large file support.  */
   {
     loff_t offset, result;
-    static int _llseek (uint fd, ulong hi, ulong lo, loff_t *res, uint wh);
-    _syscall5 (int, _llseek, uint, fd, ulong, hi, ulong, lo,
+    static int _llseek (uint filedes, ulong hi, ulong lo,
+			loff_t *res, uint wh);
+    _syscall5 (int, _llseek, uint, filedes, ulong, hi, ulong, lo,
 	       loff_t *, res, uint, wh);
 
     offset = (loff_t) sector * (loff_t) SECTOR_SIZE;
@@ -892,7 +814,7 @@ biosdisk (int subfunc, int drive, struct geometry *geometry,
 #else
   {
     off_t offset = (off_t) sector * (off_t) SECTOR_SIZE;
-    
+
     if (lseek (fd, offset, SEEK_SET) != offset)
       return -1;
   }
@@ -906,7 +828,7 @@ biosdisk (int subfunc, int drive, struct geometry *geometry,
       if (nread (fd, buf, nsec * SECTOR_SIZE) != nsec * SECTOR_SIZE)
 	return -1;
       break;
-      
+
     case BIOSDISK_WRITE:
       if (verbose)
 	{
@@ -919,12 +841,12 @@ biosdisk (int subfunc, int drive, struct geometry *geometry,
 	if (nwrite (fd, buf, nsec * SECTOR_SIZE) != nsec * SECTOR_SIZE)
 	  return -1;
       break;
-      
+
     default:
       grub_printf ("unknown subfunc %d\n", subfunc);
       break;
     }
-  
+
   return 0;
 }
 
@@ -934,3 +856,226 @@ stop_floppy (void)
 {
   /* NOTUSED */
 }
+
+/* The serial version of getkey.  */
+int
+serial_getkey (void)
+{
+  char c;
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+  struct timeval otv, tv;
+
+  gettimeofday (&otv, 0);
+#endif /* SIMULATE_SLOWNESS_OF_SERIAL */
+  
+  if (nread (serial_fd, &c, 1) != 1)
+    stop ();
+
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+  while (1)
+    {
+      long delta;
+      
+      gettimeofday (&tv, 0);
+      delta = tv.tv_usec - otv.tv_usec;
+      if (delta < 0)
+	delta += 1000000;
+      
+      if (delta >= 1000000 / (serial_speed >> 3))
+	break;
+    }
+#endif /* SIMULATE_SLOWNESS_OF_SERIAL */
+  
+  return c;
+}
+
+/* The serial version of checkkey.  */
+int
+serial_checkkey (void)
+{
+  fd_set fds;
+  struct timeval to;
+
+  /* Wait only for the serial device.  */
+  FD_ZERO (&fds);
+  FD_SET (serial_fd, &fds);
+
+  /* Set the timeout to 100ms.  */
+  to.tv_sec = 0;
+  to.tv_usec = 100000;
+  
+  return select (serial_fd + 1, &fds, 0, 0, &to) > 0 ? : -1;
+}
+
+/* The serial version of grub_putchar.  */
+void
+serial_putchar (int c)
+{
+  char ch = (char) c;
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+  struct timeval otv, tv;
+  
+  gettimeofday (&otv, 0);
+#endif /* SIMULATE_SLOWNESS_OF_SERIAL */
+  
+  if (nwrite (serial_fd, &ch, 1) != 1)
+    stop ();
+  
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+  while (1)
+    {
+      long delta;
+      
+      gettimeofday (&tv, 0);
+      delta = tv.tv_usec - otv.tv_usec;
+      if (delta < 0)
+	delta += 1000000;
+      
+      if (delta >= 1000000 / (serial_speed >> 3))
+	break;
+    }
+#endif /* SIMULATE_SLOWNESS_OF_SERIAL */
+}
+
+static speed_t
+get_termios_speed (int speed)
+{
+  switch (speed)
+    {
+    case 2400: return B2400;
+    case 4800: return B4800;
+    case 9600: return B9600;
+    case 19200: return B19200;
+    case 38400: return B38400;
+#ifdef B57600
+    case 57600: return B57600;
+#endif
+#ifdef B115200      
+    case 115200: return B115200;
+#endif
+    }
+
+  return B0;
+}
+
+/* Get the port number of the unit UNIT. In the grub shell, this doesn't
+   make sense.  */
+unsigned short
+serial_get_port (int unit)
+{
+  return 0;
+}
+
+/* Initialize a serial device. In the grub shell, PORT is unused.  */
+int
+serial_init (unsigned short port, unsigned int speed,
+	     int word_len, int parity, int stop_bit_len)
+{
+  struct termios termios;
+  speed_t termios_speed;
+  
+  /* Check if the file name is specified.  */
+  if (! serial_device)
+    return 0;
+
+  /* If a serial device is already opened, close it first.  */
+  if (serial_fd >= 0)
+    close (serial_fd);
+  
+  /* Open the device file.  */
+  serial_fd = open (serial_device, O_RDWR | O_NOCTTY | O_SYNC);
+  if (serial_fd < 0)
+    return 0;
+
+  /* Get the termios parameters.  */
+  if (tcgetattr (serial_fd, &termios))
+    goto fail;
+
+  /* Raw mode.  */
+  cfmakeraw (&termios);
+
+  /* Set the speed.  */
+  termios_speed = get_termios_speed (speed);
+  if (termios_speed == B0)
+    goto fail;
+  
+  cfsetispeed (&termios, termios_speed);
+  cfsetospeed (&termios, termios_speed);
+
+  /* Set the word length.  */
+  termios.c_cflag &= ~CSIZE;
+  switch (word_len)
+    {
+    case UART_5BITS_WORD:
+      termios.c_cflag |= CS5;
+      break;
+    case UART_6BITS_WORD:
+      termios.c_cflag |= CS6;
+      break;
+    case UART_7BITS_WORD:
+      termios.c_cflag |= CS7;
+      break;
+    case UART_8BITS_WORD:
+      termios.c_cflag |= CS8;
+      break;
+    default:
+      goto fail;
+    }
+
+  /* Set the parity.  */
+  switch (parity)
+    {
+    case UART_NO_PARITY:
+      termios.c_cflag &= ~PARENB;
+      break;
+    case UART_ODD_PARITY:
+      termios.c_cflag |= PARENB;
+      termios.c_cflag |= PARODD;
+      break;
+    case UART_EVEN_PARITY:
+      termios.c_cflag |= PARENB;
+      termios.c_cflag &= ~PARODD;
+      break;
+    default:
+      goto fail;
+    }
+
+  /* Set the length of stop bit.  */
+  switch (stop_bit_len)
+    {
+    case UART_1_STOP_BIT:
+      termios.c_cflag &= ~CSTOPB;
+      break;
+    case UART_2_STOP_BITS:
+      termios.c_cflag |= CSTOPB;
+      break;
+    default:
+      goto fail;
+    }
+
+  /* Set the parameters.  */
+  if (tcsetattr (serial_fd, TCSANOW, &termios))
+    goto fail;
+
+#ifdef SIMULATE_SLOWNESS_OF_SERIAL
+  serial_speed = speed;
+#endif /* SIMUATE_SLOWNESS_OF_SERIAL */
+  
+  return 1;
+
+ fail:
+  close (serial_fd);
+  serial_fd = -1;
+  return 0;
+}
+
+/* Set the file name of a serial device (or a pty device). This is a
+   function specific to the grub shell.  */
+void
+set_serial_device (const char *device)
+{
+  if (serial_device)
+    free (serial_device);
+  
+  serial_device = strdup (device);
+}
