@@ -1,25 +1,10 @@
 /* cs89x0.c: A Crystal Semiconductor CS89[02]0 driver for etherboot. */
 /*
-  This code is heavily based on the linux driver as written by
-  Russell Nelson <nelson@crynwr.com> and Donald Becker
-  <becker@cesdis.gsfc.nasa.gov> and modified by Mike Cruse
-  <mcruse@cti-ltd.com>. That driver has been released under the
-  conditions of the GNU Public License, thus this one is probably to
-  be considered "derived work". Therefore, there are some legal
-  obstacles in combining this file with etherboot's code, which is
-  released under a BSD style license.
-
-  So, if you want to actually use this code, you should make sure that
-  you are aware of the legal implications. I release *my* work into
-  the PUBLIC DOMAIN which implies that you can use it either under a
-  BSD style license, or under the conditions of the GPL, or under any
-  other conditions that you like. N.B. this does not apply to the
-  parts that originate from other authors. So, you should probably
-  contact them first and verify if they agree with your intended use.
-
-  If you contacted all of the above authors and they agreed to give
-  special permission for using this code under the conditions of a BSD
-  style license, then please do let me know.
+  Permission is granted to distribute the enclosed cs89x0.[ch] driver
+  only in conjunction with the Etherboot package.  The code is
+  ordinarily distributed under the GPL.
+  
+  Russ Nelson, January 2000
 
   ChangeLog:
 
@@ -34,7 +19,7 @@
   * some cosmetic changes
 
   * 2.5
-  
+
   Thu Dec 5 21:00:00 1996  Markus Gutschke  <gutschk@math.uni-muenster.de>
 
   * tested the code against a CS8900 card
@@ -45,7 +30,7 @@
     changes in order to be more tolerant to different environments
 
   * 4
-  
+
   Fri Nov 22 23:00:00 1996  Markus Gutschke  <gutschk@math.uni-muenster.de>
 
   * read the manuals for the CS89x0 chipsets and took note of all the
@@ -65,29 +50,28 @@
   * Cleaned up some of the code and tried to optimize the code size.
 
   * 1.5
-  
+
   Sun Nov 10 16:30:00 1996  Markus Gutschke  <gutschk@math.uni-muenster.de>
 
   * First experimental release. This code compiles fine, but I
   have no way of testing whether it actually works.
 
   * I did not (yet) bother to make the code 16bit aware, so for
-  the time being, it will only work for netboot-32.
+  the time being, it will only work for Etherboot/32.
 
   * 12
-  
+
   */
 
-#ifdef INCLUDE_CS
-
-#include "netboot.h"
+#include "etherboot.h"
 #include "nic.h"
+#include "cards.h"
 #include "cs89x0.h"
 
 static unsigned short	eth_nic_base;
 static unsigned long    eth_mem_start;
 static unsigned short   eth_irq;
-static unsigned short   eth_cs_type;	 /* one of: CS8900, CS8920, CS8920M  */
+static unsigned short   eth_cs_type;	/* one of: CS8900, CS8920, CS8920M  */
 static unsigned short   eth_auto_neg_cnf;
 static unsigned short   eth_adapter_cnf;
 static unsigned short	eth_linectl;
@@ -97,19 +81,16 @@ static unsigned char	eth_vendor;
 	CS89x0 - specific routines
 **************************************************************************/
 
-static int inline readreg(portno)
-	int portno;
+static inline int readreg(int portno)
 {
-	outw(eth_nic_base + ADD_PORT, portno);
+	outw(portno, eth_nic_base + ADD_PORT);
 	return inw(eth_nic_base + DATA_PORT);
 }
 
-static void inline writereg(portno,value)
-	int portno;
-	int value;
+static inline void writereg(int portno, int value)
 {
-	outw(eth_nic_base + ADD_PORT, portno);
-	outw(eth_nic_base + DATA_PORT, value);
+	outw(portno, eth_nic_base + ADD_PORT);
+	outw(value, eth_nic_base + DATA_PORT);
 	return;
 }
 
@@ -117,28 +98,25 @@ static void inline writereg(portno,value)
 EEPROM access
 **************************************************************************/
 
-static int wait_eeprom_ready()
+static int wait_eeprom_ready(void)
 {
-	long tmo = currticks() + 4*18L;
-	
+	unsigned long tmo = currticks() + 4*TICKS_PER_SEC;
+
 	/* check to see if the EEPROM is ready, a timeout is used -
 	   just in case EEPROM is ready when SI_BUSY in the
 	   PP_SelfST is clear */
 	while(readreg(PP_SelfST) & SI_BUSY) {
-		if (currticks() - tmo >= 0)
+		if (currticks() >= tmo)
 			return -1; }
 	return 0;
 }
 
-static int get_eeprom_data(off,len,buffer)
-	int off;
-	int len;
-	unsigned short *buffer;
+static int get_eeprom_data(int off, int len, unsigned short *buffer)
 {
 	int i;
 
-#if defined(EDEBUG)
-	printf("\r\ncs: EEPROM data from %x for %x:",off,len);
+#ifdef	EDEBUG
+	printf("\ncs: EEPROM data from %x for %x:",off,len);
 #endif
 	for (i = 0; i < len; i++) {
 		if (wait_eeprom_ready() < 0)
@@ -149,23 +127,20 @@ static int get_eeprom_data(off,len,buffer)
 		if (wait_eeprom_ready() < 0)
 			return -1;
 		buffer[i] = readreg(PP_EEData);
-#if defined(EDEBUG)
+#ifdef	EDEBUG
 		if (!(i%10))
-			printf("\r\ncs: ");
+			printf("\ncs: ");
 		printf("%x ", buffer[i]);
 #endif
 	}
-#if defined(EDEBUG)
-	printf("\r\n");
+#ifdef	EDEBUG
+	putchar('\n');
 #endif
 
 	return(0);
 }
 
-static int get_eeprom_chksum(off, len, buffer)
-	int off;
-	int len;
-	unsigned short *buffer;
+static int get_eeprom_chksum(int off, int len, unsigned short *buffer)
 {
 	int  i, cksum;
 
@@ -182,7 +157,7 @@ static int get_eeprom_chksum(off, len, buffer)
 Activate all of the available media and probe for network
 **************************************************************************/
 
-static void clrline()
+static void clrline(void)
 {
 	int i;
 
@@ -192,11 +167,10 @@ static void clrline()
 	return;
 }
 
-static void control_dc_dc(on_not_off)
-	int on_not_off;
+static void control_dc_dc(int on_not_off)
 {
 	unsigned int selfcontrol;
-	long tmo = currticks() + 18;
+	unsigned long tmo = currticks() + TICKS_PER_SEC;
 
 	/* control the DC to DC convertor in the SelfControl register.  */
 	selfcontrol = HCB1_ENBL; /* Enable the HCB1 bit as an output */
@@ -205,19 +179,19 @@ static void control_dc_dc(on_not_off)
 	else
 		selfcontrol &= ~HCB1;
 	writereg(PP_SelfCTL, selfcontrol);
-  
+
 	/* Wait for the DC/DC converter to power up - 1000ms */
-	while (currticks() - tmo < 0);
+	while (currticks() < tmo);
 
 	return;
 }
 
-static int detect_tp()
+static int detect_tp(void)
 {
-	long tmo;
+	unsigned long tmo;
 
 	/* Turn on the chip auto detection of 10BT/ AUI */
-  
+
 	clrline(); printf("attempting %s:","TP");
 
         /* If connected to another full duplex capable 10-Base-T card
@@ -231,8 +205,8 @@ static int detect_tp()
 
         /* Delay for the hardware to work out if the TP cable is
 	   present - 150ms */
-	for (tmo = currticks() + 4; currticks() - tmo < 0; );
-	
+	for (tmo = currticks() + 4; currticks() < tmo; );
+
 	if ((readreg(PP_LineST) & LINK_OK) == 0)
 		return 0;
 
@@ -243,7 +217,7 @@ static int detect_tp()
 		if ((eth_auto_neg_cnf & AUTO_NEG_BITS) == AUTO_NEG_ENABLE) {
 			printf(" negotiating duplex... ");
 			while (readreg(PP_AutoNegST) & AUTO_NEG_BUSY) {
-				if (currticks() - tmo > 40*18) {
+				if (currticks() - tmo > 40*TICKS_PER_SEC) {
 					printf("time out ");
 					break;
 				}
@@ -265,21 +239,20 @@ static int send_test_pkt(struct nic *nic)
 				     0, 46, /*A 46 in network order       */
 				     0, 0,  /*DSAP=0 & SSAP=0 fields      */
 				     0xf3,0 /*Control (Test Req+P bit set)*/ };
-	long tmo;
+	unsigned long tmo;
 
 	writereg(PP_LineCTL, readreg(PP_LineCTL) | SERIAL_TX_ON);
 
-	bcopy(nic->node_addr,testpacket,ETHER_ADDR_SIZE);
-	bcopy(nic->node_addr,testpacket+ETHER_ADDR_SIZE,
-	      ETHER_ADDR_SIZE);
+	memcpy(testpacket, nic->node_addr, ETHER_ADDR_SIZE);
+	memcpy(testpacket+ETHER_ADDR_SIZE, nic->node_addr, ETHER_ADDR_SIZE);
 
-	outw(eth_nic_base + TX_CMD_PORT, TX_AFTER_ALL);
-	outw(eth_nic_base + TX_LEN_PORT, ETH_MIN_PACKET);
+	outw(TX_AFTER_ALL, eth_nic_base + TX_CMD_PORT);
+	outw(ETH_MIN_PACKET, eth_nic_base + TX_LEN_PORT);
 
 	/* Test to see if the chip has allocated memory for the packet */
 	for (tmo = currticks() + 2;
 	     (readreg(PP_BusST) & READY_FOR_TX_NOW) == 0; )
-		if (currticks() - tmo >= 0)
+		if (currticks() >= tmo)
 			return(0);
 
 	/* Write the contents of the packet */
@@ -288,8 +261,8 @@ static int send_test_pkt(struct nic *nic)
 
 	printf(" sending test packet ");
 	/* wait a couple of timer ticks for packet to be received */
-	for (tmo = currticks() + 2; currticks() - tmo < 0; );
-	
+	for (tmo = currticks() + 2; currticks() < tmo; );
+
 	if ((readreg(PP_TxEvent) & TX_SEND_OK_BITS) == TX_OK) {
 			printf("succeeded");
 			return 1;
@@ -329,10 +302,10 @@ static int detect_bnc(struct nic *nic)
 ETH_RESET - Reset adapter
 ***************************************************************************/
 
-void cs89x0_reset(struct nic *nic)
+static void cs89x0_reset(struct nic *nic)
 {
 	int  i;
-	long reset_tmo;
+	unsigned long reset_tmo;
 
 	if(eth_vendor!=VENDOR_CS89x0)
 		return;
@@ -340,26 +313,25 @@ void cs89x0_reset(struct nic *nic)
 	writereg(PP_SelfCTL, readreg(PP_SelfCTL) | POWER_ON_RESET);
 
 	/* wait for two ticks; that is 2*55ms */
-	for (reset_tmo = currticks() + 2;
-	     reset_tmo > currticks(); );
+	for (reset_tmo = currticks() + 2; currticks() < reset_tmo; );
 
 	if (eth_cs_type != CS8900) {
 		/* Hardware problem requires PNP registers to be reconfigured
 		   after a reset */
 		if (eth_irq != 0xFFFF) {
-			outw(eth_nic_base + ADD_PORT, PP_CS8920_ISAINT);
-			outb(eth_nic_base + DATA_PORT, eth_irq);
-			outb(eth_nic_base + DATA_PORT + 1, 0); }
+			outw(PP_CS8920_ISAINT, eth_nic_base + ADD_PORT);
+			outb(eth_irq, eth_nic_base + DATA_PORT);
+			outb(0, eth_nic_base + DATA_PORT + 1); }
 
 		if (eth_mem_start) {
-			outw(eth_nic_base + ADD_PORT, PP_CS8920_ISAMemB);
-			outb(eth_nic_base + DATA_PORT, (eth_mem_start >> 8) & 0xff);
-			outb(eth_nic_base + DATA_PORT + 1, (eth_mem_start >> 24) & 0xff); } }
-  
+			outw(PP_CS8920_ISAMemB, eth_nic_base + ADD_PORT);
+			outb((eth_mem_start >> 8) & 0xff, eth_nic_base + DATA_PORT);
+			outb((eth_mem_start >> 24) & 0xff, eth_nic_base + DATA_PORT + 1); } }
+
 	/* Wait until the chip is reset */
 	for (reset_tmo = currticks() + 2;
 	     (readreg(PP_SelfST) & INIT_DONE) == 0 &&
-		     currticks() - reset_tmo < 0; );
+		     currticks() < reset_tmo; );
 
 	/* disable interrupts and memory accesses */
 	writereg(PP_BusCTL, 0);
@@ -383,7 +355,7 @@ void cs89x0_reset(struct nic *nic)
 	writereg(PP_BufCFG, 0);
 
 	/* reset address port, so that autoprobing will keep working */
-	outw(eth_nic_base + ADD_PORT, PP_ChipID);
+	outw(PP_ChipID, eth_nic_base + ADD_PORT);
 
 	return;
 }
@@ -394,12 +366,12 @@ ETH_TRANSMIT - Transmit a frame
 
 static void cs89x0_transmit(
 	struct nic *nic,
-	char *d,			/* Destination */
+	const char *d,			/* Destination */
 	unsigned int t,			/* Type */
 	unsigned int s,			/* size */
-	char *p)			/* Packet */
+	const char *p)			/* Packet */
 {
-	long          tmo;
+	unsigned long tmo;
 	int           sr;
 
 	if(eth_vendor!=VENDOR_CS89x0)
@@ -412,14 +384,14 @@ static void cs89x0_transmit(
 
 retry:
 	/* initiate a transmit sequence */
-	outw(eth_nic_base + TX_CMD_PORT, TX_AFTER_ALL);
-	outw(eth_nic_base + TX_LEN_PORT, sr);
+	outw(TX_AFTER_ALL, eth_nic_base + TX_CMD_PORT);
+	outw(sr, eth_nic_base + TX_LEN_PORT);
 
 	/* Test to see if the chip has allocated memory for the packet */
 	if ((readreg(PP_BusST) & READY_FOR_TX_NOW) == 0) {
 		/* Oops... this should not happen! */
-		printf("cs: unable to send packet; retrying...\r\n");
-		for (tmo = currticks() + 5*18; currticks() - tmo < 0; );
+		printf("cs: unable to send packet; retrying...\n");
+		for (tmo = currticks() + 5*TICKS_PER_SEC; currticks() < tmo; );
 		cs89x0_reset(nic);
 		goto retry; }
 
@@ -427,17 +399,17 @@ retry:
 	outsw(eth_nic_base + TX_FRAME_PORT, d, ETHER_ADDR_SIZE/2);
 	outsw(eth_nic_base + TX_FRAME_PORT, nic->node_addr,
 	      ETHER_ADDR_SIZE/2);
-	outw(eth_nic_base + TX_FRAME_PORT, ((t >> 8)&0xFF)|(t << 8));
+	outw(((t >> 8)&0xFF)|(t << 8), eth_nic_base + TX_FRAME_PORT);
 	outsw(eth_nic_base + TX_FRAME_PORT, p, (s+1)/2);
 	for (sr = sr/2 - (s+1)/2 - ETHER_ADDR_SIZE - 1; sr-- > 0;
-	     outw(eth_nic_base + TX_FRAME_PORT, 0));
+	     outw(0, eth_nic_base + TX_FRAME_PORT));
 
 	/* wait for transfer to succeed */
-	for (tmo = currticks()+5*18;
-	     (s = readreg(PP_TxEvent)&~0x1F) == 0 && currticks() - tmo < 0;
-	     twiddle());
+	for (tmo = currticks()+5*TICKS_PER_SEC;
+	     (s = readreg(PP_TxEvent)&~0x1F) == 0 && currticks() < tmo;)
+		/* nothing */ ;
 	if ((s & TX_SEND_OK_BITS) != TX_OK) {
-		printf("\r\ntransmission error 0x%x\r\n", s);
+		printf("\ntransmission error 0x%x\n", s);
 	}
 
 	return;
@@ -478,12 +450,12 @@ ETH_PROBE - Look for an adapter
 struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 {
 	static const unsigned int netcard_portlist[] = {
-#ifdef CS_SCAN
+#ifdef	CS_SCAN
 		CS_SCAN,
-#else /* use "conservative" default values for autoprobing */
+#else	/* use "conservative" default values for autoprobing */
 		0x300,0x320,0x340,0x200,0x220,0x240,
 		0x260,0x280,0x2a0,0x2c0,0x2e0,
-      /* if that did not work, then be more aggressive */
+	/* if that did not work, then be more aggressive */
 		0x301,0x321,0x341,0x201,0x221,0x241,
 		0x261,0x281,0x2a1,0x2c1,0x2e1,
 #endif
@@ -504,7 +476,7 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			ioaddr &= ~1;
 			if ((inw(ioaddr + ADD_PORT) & ADD_MASK) != ADD_SIG)
 				continue;
-			outw(ioaddr + ADD_PORT, PP_ChipID);
+			outw(PP_ChipID, ioaddr + ADD_PORT);
 		}
 
 		if (inw(ioaddr + DATA_PORT) != CHIP_EISA_ID_SIG)
@@ -516,26 +488,26 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 		eth_cs_type = rev_type &~ REVISON_BITS;
 		cs_revision = ((rev_type & REVISON_BITS) >> 8) + 'A';
 
-		printf("\r\ncs: cs89%c0%s rev %c, base 0x%x",
+		printf("\ncs: cs89%c0%s rev %c, base 0x%x",
 		       eth_cs_type==CS8900?'0':'2',
 		       eth_cs_type==CS8920M?"M":"",
 		       cs_revision,
 		       eth_nic_base);
-    
+
 		/* First check to see if an EEPROM is attached*/
 		if ((readreg(PP_SelfST) & EEPROM_PRESENT) == 0) {
-			printf("\r\ncs: no EEPROM...\r\n");
-			outw(eth_nic_base + ADD_PORT, PP_ChipID);
+			printf("\ncs: no EEPROM...\n");
+			outw(PP_ChipID, eth_nic_base + ADD_PORT);
 			continue; }
 		else if (get_eeprom_data(START_EEPROM_DATA,CHKSUM_LEN,
 					 eeprom_buff) < 0) {
-			printf("\r\ncs: EEPROM read failed...\r\n");
-			outw(eth_nic_base + ADD_PORT, PP_ChipID);
+			printf("\ncs: EEPROM read failed...\n");
+			outw(PP_ChipID, eth_nic_base + ADD_PORT);
 			continue; }
 		else if (get_eeprom_chksum(START_EEPROM_DATA,CHKSUM_LEN,
 					   eeprom_buff) < 0) {
-			printf("\r\ncs: EEPROM checksum bad...\r\n");
-			outw(eth_nic_base + ADD_PORT, PP_ChipID);
+			printf("\ncs: EEPROM checksum bad...\n");
+			outw(PP_ChipID, eth_nic_base + ADD_PORT);
 			continue; }
 
 		/* get transmission control word but keep the
@@ -548,12 +520,12 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 
 		/* store the initial memory base address */
 		eth_mem_start = eeprom_buff[PACKET_PAGE_OFFSET/2] << 8;
-    
+
 		printf("%s%s%s, addr ",
 		       (eth_adapter_cnf & A_CNF_10B_T)?", RJ-45":"",
 		       (eth_adapter_cnf & A_CNF_AUI)?", AUI":"",
 		       (eth_adapter_cnf & A_CNF_10B_2)?", BNC":"");
-  
+
 		/* If this is a CS8900 then no pnp soft */
 		if (eth_cs_type != CS8900 &&
 		    /* Check if the ISA IRQ has been set  */
@@ -572,14 +544,14 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 				   write_irq, which is the reverse
 				   mapping of the table below. */
 				if (i < 4) i = "\012\013\014\005"[i];
-				else printf("\r\ncs: BUG: isa_config is %d\r\n", i); }
+				else printf("\ncs: BUG: isa_config is %d\n", i); }
 			eth_irq = i; }
 
 		/* Retrieve and print the ethernet address. */
-    		for (i=0; i<ETHER_ADDR_SIZE; i++) {
+		for (i=0; i<ETHER_ADDR_SIZE; i++) {
 			printf("%b%s",(int)(nic->node_addr[i] =
 					  ((unsigned char *)eeprom_buff)[i]),
-			       i < ETHER_ADDR_SIZE-1 ? ":" : "\r\n"); }
+			       i < ETHER_ADDR_SIZE-1 ? ":" : "\n"); }
 
 		/* Set the LineCTL quintuplet based on adapter
 		   configuration read from EEPROM */
@@ -598,15 +570,15 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			break;
 		case A_CNF_MEDIA_10B_2: result = eth_adapter_cnf & A_CNF_10B_2;
 			break;
-		default: result = eth_adapter_cnf & (A_CNF_10B_T | A_CNF_AUI | 
+		default: result = eth_adapter_cnf & (A_CNF_10B_T | A_CNF_AUI |
 						     A_CNF_10B_2);
 		}
 		if (!result) {
-			printf("cs: EEPROM is configured for unavailable media\r\n");
+			printf("cs: EEPROM is configured for unavailable media\n");
 		error:
-			writereg(PP_LineCTL, readreg(PP_LineCTL) & 
+			writereg(PP_LineCTL, readreg(PP_LineCTL) &
 				 ~(SERIAL_TX_ON | SERIAL_RX_ON));
-			outw(eth_nic_base + ADD_PORT, PP_ChipID);
+			outw(PP_ChipID, eth_nic_base + ADD_PORT);
 			continue;
 		}
 
@@ -614,7 +586,7 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 		eth_vendor = VENDOR_CS89x0;
 		cs89x0_reset(nic);
 		eth_vendor = VENDOR_NONE;
-		
+
 		/* set the hardware to the configured choice */
 		switch(eth_adapter_cnf & A_CNF_MEDIA_TYPE) {
 		case A_CNF_MEDIA_10B_T:
@@ -622,7 +594,7 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			if (!result) {
 				clrline();
 				printf("10Base-T (RJ-45%s",
-				       ") has no cable\r\n"); }
+				       ") has no cable\n"); }
 			/* check "ignore missing media" bit */
 			if (eth_auto_neg_cnf & IMM_BIT)
 				/* Yes! I don't care if I see a link pulse */
@@ -633,7 +605,7 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			if (!result) {
 				clrline();
 				printf("10Base-5 (AUI%s",
-				       ") has no cable\r\n"); }
+				       ") has no cable\n"); }
 			/* check "ignore missing media" bit */
 			if (eth_auto_neg_cnf & IMM_BIT)
 				/* Yes! I don't care if I see a carrrier */
@@ -644,7 +616,7 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			if (!result) {
 				clrline();
 				printf("10Base-2 (BNC%s",
-				       ") has no cable\r\n"); }
+				       ") has no cable\n"); }
 			/* check "ignore missing media" bit */
 			if (eth_auto_neg_cnf & IMM_BIT)
 				/* Yes! I don't care if I can xmit a packet */
@@ -661,18 +633,18 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 			if (eth_adapter_cnf & A_CNF_10B_2)
 				if ((result = detect_bnc(nic)) != 0)
 					break;
-			clrline(); printf("no media detected\r\n");
+			clrline(); printf("no media detected\n");
 			goto error;
 		}
 		clrline();
 		switch(result) {
-		case 0:                 printf("no network cable attached to configured media\r\n");
+		case 0:                 printf("no network cable attached to configured media\n");
 			goto error;
-		case A_CNF_MEDIA_10B_T: printf("using 10Base-T (RJ-45)\r\n");
+		case A_CNF_MEDIA_10B_T: printf("using 10Base-T (RJ-45)\n");
 			break;
-		case A_CNF_MEDIA_AUI:   printf("using 10Base-5 (AUI)\r\n");
+		case A_CNF_MEDIA_AUI:   printf("using 10Base-5 (AUI)\n");
 			break;
-		case A_CNF_MEDIA_10B_2: printf("using 10Base-2 (BNC)\r\n");
+		case A_CNF_MEDIA_10B_2: printf("using 10Base-2 (BNC)\n");
 			break;
 		}
 
@@ -690,10 +662,9 @@ struct nic *cs89x0_probe(struct nic *nic, unsigned short *probe_addrs)
 	return nic;
 }
 
-#endif /* INCLUDE_CS */
-
 /*
  * Local variables:
  *  c-basic-offset: 8
  * End:
  */
+
