@@ -2,7 +2,7 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
  *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2000  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -38,12 +38,14 @@ static struct mod_list mll[99];
  */
 
 kernel_t
-load_image (char *kernel, char *arg)
+load_image (char *kernel, char *arg, kernel_t suggested_type,
+	    unsigned long load_flags)
 {
   int len, i, exec_type = 0, align_4k = 1;
   kernel_t type = KERNEL_TYPE_NONE;
   unsigned long flags = 0, text_len = 0, data_len = 0, bss_len = 0;
   char *str = 0, *str2 = 0;
+  struct linux_kernel_header *lh;
   union
     {
       struct multiboot_header *mb;
@@ -64,6 +66,8 @@ load_image (char *kernel, char *arg)
 
   if (!(len = grub_read (buffer, MULTIBOOT_SEARCH)) || len < 32)
     {
+      grub_close ();
+      
       if (!errnum)
 	errnum = ERR_EXEC_FORMAT;
 
@@ -77,6 +81,7 @@ load_image (char *kernel, char *arg)
 	  flags = ((struct multiboot_header *) (buffer + i))->flags;
 	  if (flags & MULTIBOOT_UNSUPPORTED)
 	    {
+	      grub_close ();
 	      errnum = ERR_BOOT_FEATURES;
 	      return KERNEL_TYPE_NONE;
 	    }
@@ -86,9 +91,14 @@ load_image (char *kernel, char *arg)
 	}
     }
 
-  /* ELF loading supported if multiboot and FreeBSD.  */
+  /* Use BUFFER as a linux kernel header, if the image is Linux zImage
+     or bzImage.  */
+  lh = (struct linux_kernel_header *) buffer;
+  
+  /* ELF loading supported if multiboot, FreeBSD and NetBSD.  */
   if ((type == KERNEL_TYPE_MULTIBOOT
-       || grub_strcmp (pu.elf->e_ident + EI_BRAND, "FreeBSD") == 0)
+       || grub_strcmp (pu.elf->e_ident + EI_BRAND, "FreeBSD") == 0
+       || suggested_type == KERNEL_TYPE_NETBSD)
       && len > sizeof (Elf32_Ehdr)
       && BOOTABLE_I386_ELF ((*((Elf32_Ehdr *) buffer))))
     {
@@ -110,8 +120,18 @@ load_image (char *kernel, char *arg)
 
       if (type == KERNEL_TYPE_NONE)
 	{
-	  str2 = "FreeBSD";
-	  type = KERNEL_TYPE_FREEBSD;
+	  /* At the moment, there is no way to identify a NetBSD ELF
+	     kernel, so rely on the suggested type by the user.  */
+	  if (suggested_type == KERNEL_TYPE_NETBSD)
+	    {
+	      str2 = "NetBSD";
+	      type = suggested_type;
+	    }
+	  else
+	    {
+	      str2 = "FreeBSD";
+	      type = KERNEL_TYPE_FREEBSD;
+	    }
 	}
     }
   else if (flags & MULTIBOOT_AOUT_KLUDGE)
@@ -120,7 +140,7 @@ load_image (char *kernel, char *arg)
       entry_addr = (entry_func) pu.mb->entry_addr;
       cur_addr = pu.mb->load_addr;
       /* first offset into file */
-      filepos = i - (pu.mb->header_addr - cur_addr);
+      grub_seek (i - (pu.mb->header_addr - cur_addr));
       text_len = pu.mb->load_end_addr - cur_addr;
       data_len = 0;
       bss_len = pu.mb->bss_end_addr - pu.mb->load_end_addr;
@@ -155,25 +175,26 @@ load_image (char *kernel, char *arg)
 	   *  one...  plus of course we need to know which booting
 	   *  method to use.
 	   */
+	  entry_addr = (entry_func) ((int) entry_addr & 0xFFFFFF);
+	  
 	  if (buffer[0] == 0xb && buffer[1] == 1)
 	    {
 	      type = KERNEL_TYPE_FREEBSD;
-	      entry_addr = (entry_func) (((int) entry_addr) & 0xFFFFFF);
+	      cur_addr = (int) entry_addr;
 	      str2 = "FreeBSD";
 	    }
 	  else
 	    {
 	      type = KERNEL_TYPE_NETBSD;
-	      entry_addr = (entry_func) (((int) entry_addr) & 0xF00000);
+	      cur_addr = (int) entry_addr & 0xF00000;
 	      if (N_GETMAGIC ((*(pu.aout))) != NMAGIC)
 		align_4k = 0;
 	      str2 = "NetBSD";
 	    }
 	}
 
-      cur_addr = (int) entry_addr;
       /* first offset into file */
-      filepos = N_TXTOFF ((*(pu.aout)));
+      grub_seek (N_TXTOFF (*(pu.aout)));
       text_len = pu.aout->a_text;
       data_len = pu.aout->a_data;
       bss_len = pu.aout->a_bss;
@@ -184,32 +205,69 @@ load_image (char *kernel, char *arg)
       exec_type = 1;
       str = "a.out";
     }
-  else if ((*((unsigned short *) (buffer + BOOTSEC_SIG_OFFSET))
-	    == BOOTSEC_SIGNATURE)
-	   && ((data_len
-		= (((long) *((unsigned char *)
-			     (buffer + LINUX_SETUP_LEN_OFFSET))) << 9))
-	       <= LINUX_SETUP_MAXLEN)
-	   && ((text_len
-		= (((long) *((unsigned short *)
-			     (buffer + LINUX_KERNEL_LEN_OFFSET))) << 4)),
-	       (data_len + text_len + SECTOR_SIZE) <= ((filemax + 15) & 0xFFFFFFF0)))
+  else if (lh->boot_flag == BOOTSEC_SIGNATURE
+	   && lh->setup_sects <= LINUX_MAX_SETUP_SECTS)
     {
-      int big_linux = buffer[LINUX_SETUP_LOAD_FLAGS] & LINUX_FLAG_BIG_KERNEL;
-      buffer[LINUX_SETUP_LOADER] = 0x70;
-      if (!big_linux && text_len > LINUX_KERNEL_MAXLEN)
+      int big_linux = 0;
+      int setup_sects = lh->setup_sects;
+
+      if (lh->header == LINUX_MAGIC_SIGNATURE && lh->version >= 0x0200)
 	{
-	  printf (" linux 'zImage' kernel too big, try 'make bzImage'\n");
+	  big_linux = (lh->loadflags & LINUX_FLAG_BIG_KERNEL);
+	  lh->type_of_loader = LINUX_BOOT_LOADER_TYPE;
+
+	  if (lh->version >= 0x0201)
+	    {
+	      lh->heap_end_ptr = LINUX_HEAP_END_OFFSET;
+	      lh->loadflags |= LINUX_FLAG_CAN_USE_HEAP;
+	    }
+
+	  if (lh->version >= 0x0202)
+	    lh->cmd_line_ptr = CL_MY_LOCATION;
+	  else
+	    {
+	      lh->cl_magic = CL_MAGIC;
+	      lh->cl_offset = CL_MY_LOCATION - CL_BASE_ADDR;
+	      lh->setup_move_size
+		= (unsigned short) (CL_MY_END_ADDR - CL_BASE_ADDR + 1);
+	    }
+	}
+      else
+	{
+	  /* Your kernel is quite old...  */
+	  lh->cl_magic = CL_MAGIC;
+	  lh->cl_offset = CL_MY_LOCATION - CL_BASE_ADDR;
+	  
+	  setup_sects = LINUX_DEFAULT_SETUP_SECTS;
+	}
+      
+      /* If SETUP_SECTS is not set, set it to the default (4).  */
+      if (! setup_sects)
+	setup_sects = LINUX_DEFAULT_SETUP_SECTS;
+
+      data_len = setup_sects << 9;
+      text_len = filemax - data_len - SECTOR_SIZE;
+      
+      if (! big_linux && text_len > LINUX_KERNEL_MAXLEN)
+	{
+	  grub_printf (" linux 'zImage' kernel too big, try 'make bzImage'\n");
+	  grub_close ();
 	  errnum = ERR_WONT_FIT;
 	  return KERNEL_TYPE_NONE;
 	}
 
-      printf ("   [Linux-%s, setup=0x%x, size=0x%x]\n",
-	      (big_linux ? "bzImage" : "zImage"), data_len, text_len);
+      grub_printf ("   [Linux-%s, setup=0x%x, size=0x%x]\n",
+		   (big_linux ? "bzImage" : "zImage"), data_len, text_len);
 
-      if (mbi.mem_lower >= 608)
+      /* FIXME: SETUP_SECTS should be supported up to 63.
+	 But do you know there are >640KB conventional memory machines?  */
+      if (mbi.mem_lower >= 608 && setup_sects < 60)
 	{
-	  /* Video mode selection support. What a shit!  */
+	  /* Video mode selection support. What a mess!  */
+	  /* NOTE: Even the word "mess" is not still enough to
+	     represent how wrong and bad the Linux video support is,
+	     but I don't want to hear complaints from Linux fanatics
+	     any more. -okuji  */
 	  {
 	    char *vga;
 
@@ -218,7 +276,6 @@ load_image (char *kernel, char *arg)
 	    if (vga)
 	      {
 		char *value = vga + 4;
-		char *vga_end;
 		int vid_mode;
 
 		/* Handle special strings.  */
@@ -231,53 +288,83 @@ load_image (char *kernel, char *arg)
 		else if (safe_parse_maxint (&value, &vid_mode))
 		  ;
 		else
-		  /* ERRNUM is already set inside the function
-		     safe_parse_maxint.  */
-		  return KERNEL_TYPE_NONE;
+		  {
+		    /* ERRNUM is already set inside the function
+		       safe_parse_maxint.  */
+		    grub_close ();
+		    return KERNEL_TYPE_NONE;
+		  }
 
-		/* Set the vid mode to VID_MODE. Note that this can work
-		   because i386 architecture is little-endian.  */
-		grub_memmove (buffer + LINUX_VID_MODE_OFFSET,
-			      (char *) &vid_mode,
-			      sizeof (unsigned short));
-		
-		/* Remove the "vga=...".  */
-		vga_end = skip_to (0, vga);
-		grub_memmove (vga, vga_end, grub_strlen (vga_end) + 1);
+		lh->vid_mode = vid_mode;
 	      }
 	  }
 		
 	  memmove ((char *) LINUX_SETUP, buffer, data_len + SECTOR_SIZE);
 
-	  /* copy command-line plus memory hack to staging area */
+	  if (lh->header != LINUX_MAGIC_SIGNATURE ||
+	      lh->version < 0x0200)
+	    /* Clear the heap space.  */
+	    grub_memset ((char *) LINUX_SETUP + ((setup_sects - 1) << 9),
+			 0,
+			 (64 - setup_sects - 1) << 9);
+
+	  /* Copy command-line plus memory hack to staging area.
+	     NOTE: Linux has a bug that it doesn't handle multiple spaces
+	     between two options and a space after a "mem=" option isn't
+	     removed correctly so the arguments to init could be like
+	     {"init", "", "", NULL}. This affects some not-very-clever
+	     shells. Thus, the code below does a trick to avoid the bug.
+	     That is, copy "mem=XXX" to the end of the command-line, and
+	     avoid to copy spaces unnecessarily. Hell.  */
 	  {
-	    char *src = arg;
-	    char *dest = (char *) (CL_MY_LOCATION + 4);
+	    char *src = skip_to (0, arg);
+	    char *dest = (char *) CL_MY_LOCATION;
 
-	    memmove ((char *) CL_MY_LOCATION, "mem=", 4);
-
-	    *((unsigned short *) CL_OFFSET) = CL_MY_LOCATION - CL_BASE_ADDR;
-	    *((unsigned short *) CL_MAGIC_ADDR) = CL_MAGIC;
-
-	    dest = convert_to_ascii (dest, 'u', (mbi.mem_upper + 0x400));
-	    *(dest++) = 'K';
-	    *(dest++) = ' ';
-
-	    while (*src && *src != ' ')
-	      src++;
-
-	    while (((int) dest) < CL_MY_END_ADDR && (*(dest++) = *(src++)));
+	    while (((int) dest) < CL_MY_END_ADDR && *src)
+	      *(dest++) = *(src++);
+	    
+	    /* Add a mem option automatically only if the user doesn't
+	       specify it explicitly.  */
+	    if (! grub_strstr (arg, "mem=")
+		&& ! (load_flags & KERNEL_LOAD_NO_MEM_OPTION))
+	      {
+		if (dest != (char *) CL_MY_LOCATION)
+		  *(dest++) = ' ';
+		
+		grub_memmove (dest, "mem=", 4);
+		dest += 4;
+		
+		dest = convert_to_ascii (dest, 'u', (extended_memory + 0x400));
+		*(dest++) = 'K';
+	      }
 
 	    *dest = 0;
 	  }
 
 	  /* offset into file */
-	  filepos = data_len + SECTOR_SIZE;
+	  grub_seek (data_len + SECTOR_SIZE);
 
 	  cur_addr = LINUX_STAGING_AREA + text_len;
-	  if (grub_read ((char *) LINUX_STAGING_AREA, text_len) >= (text_len - 16))
-	    return (big_linux ? KERNEL_TYPE_BIG_LINUX : KERNEL_TYPE_LINUX);
-	  else if (!errnum)
+	  if (grub_read ((char *) LINUX_STAGING_AREA, text_len)
+	      >= (text_len - 16))
+	    {
+	      grub_close ();
+
+	      /* Sanity check.  */
+	      if (suggested_type != KERNEL_TYPE_NONE
+		  && ((big_linux && suggested_type != KERNEL_TYPE_BIG_LINUX)
+		      || (! big_linux && suggested_type != KERNEL_TYPE_LINUX)))
+		{
+		  errnum = ERR_EXEC_FORMAT;
+		  return KERNEL_TYPE_NONE;
+		}
+
+	      /* Ugly hack.  */
+	      linux_text_len = text_len;
+	      
+	      return big_linux ? KERNEL_TYPE_BIG_LINUX : KERNEL_TYPE_LINUX;
+	    }
+	  else if (! errnum)
 	    errnum = ERR_EXEC_FORMAT;
 	}
       else
@@ -288,7 +375,10 @@ load_image (char *kernel, char *arg)
 
   /* return if error */
   if (errnum)
-    return KERNEL_TYPE_NONE;
+    {
+      grub_close ();
+      return KERNEL_TYPE_NONE;
+    }
 
   /* fill the multiboot info structure */
   mbi.cmdline = (int) arg;
@@ -351,7 +441,8 @@ load_image (char *kernel, char *arg)
 	  int symtab_err, orig_addr = cur_addr;
 
 	  /* we should align to a 4K boundary here for good measure */
-	  cur_addr = (cur_addr + 0xFFF) & 0xFFFFF000;
+	  if (align_4k)
+	    cur_addr = (cur_addr + 0xFFF) & 0xFFFFF000;
 
 	  mbi.syms.a.addr = cur_addr;
 
@@ -417,10 +508,10 @@ load_image (char *kernel, char *arg)
 	  if (phdr->p_type == PT_LOAD)
 	    {
 	      /* offset into file */
-	      filepos = phdr->p_offset;
+	      grub_seek (phdr->p_offset);
 	      filesiz = phdr->p_filesz;
 	      
-	      if (type == KERNEL_TYPE_FREEBSD)
+	      if (type == KERNEL_TYPE_FREEBSD || type == KERNEL_TYPE_NETBSD)
 		memaddr = RAW_ADDR (phdr->p_paddr & 0xFFFFFF);
 	      else
 		memaddr = RAW_ADDR (phdr->p_paddr);
@@ -470,6 +561,15 @@ load_image (char *kernel, char *arg)
       type = KERNEL_TYPE_NONE;
     }
 
+  grub_close ();
+
+  /* Sanity check.  */
+  if (suggested_type != KERNEL_TYPE_NONE && suggested_type != type)
+    {
+      errnum = ERR_EXEC_FORMAT;
+      return KERNEL_TYPE_NONE;
+    }
+  
   return type;
 }
 
@@ -481,9 +581,16 @@ load_module (char *module, char *arg)
   /* if we are supposed to load on 4K boundaries */
   cur_addr = (cur_addr + 0xFFF) & 0xFFFFF000;
 
-  if (!grub_open (module) || !(len = grub_read ((char *) cur_addr, -1)))
+  if (!grub_open (module))
     return 0;
 
+  len = grub_read ((char *) cur_addr, -1);
+  if (! len)
+    {
+      grub_close ();
+      return 0;
+    }
+
   printf ("   [Multiboot-module @ 0x%x, 0x%x bytes]\n", cur_addr, len);
 
   /* these two simply need to be set if any modules are loaded at all */
@@ -499,6 +606,7 @@ load_module (char *module, char *arg)
   /* increment number of modules included */
   mbi.mods_count++;
 
+  grub_close ();
   return 1;
 }
 
@@ -506,20 +614,36 @@ int
 load_initrd (char *initrd)
 {
   int len;
-  long *ramdisk, moveto;
-
-  if (! grub_open (initrd) || ! (len = grub_read ((char *) cur_addr, -1)))
+  unsigned long moveto;
+  struct linux_kernel_header *lh;
+  
+  if (! grub_open (initrd))
     return 0;
 
+  len = grub_read ((char *) cur_addr, -1);
+  if (! len)
+    {
+      grub_close ();
+      return 0;
+    }
+
   moveto = ((mbi.mem_upper + 0x400) * 0x400 - len) & 0xfffff000;
-  memmove ((void *) moveto, (void *) cur_addr, len);
+  if (moveto + len >= LINUX_INITRD_MAX_ADDRESS)
+    moveto = (LINUX_INITRD_MAX_ADDRESS - len) & 0xfffff000;
+  
+  /* XXX: Linux 2.3.xx has a bug in the memory range check, so avoid
+     the last page.  */
+  moveto -= 0x1000;
+  memmove ((void *) RAW_ADDR (moveto), (void *) cur_addr, len);
 
   printf ("   [Linux-initrd @ 0x%x, 0x%x bytes]\n", moveto, len);
 
-  ramdisk = (long *) (LINUX_SETUP + LINUX_SETUP_INITRD);
-  ramdisk[0] = moveto;
-  ramdisk[1] = len;
+  /* FIXME: Should check if the kernel supports INITRD.  */
+  lh = (struct linux_kernel_header *) LINUX_SETUP;
+  lh->ramdisk_image = RAW_ADDR (moveto);
+  lh->ramdisk_size = len;
 
+  grub_close ();
   return 1;
 }
 
@@ -623,7 +747,7 @@ bsd_boot (kernel_t type, int bootdev, char *arg)
       bi.bi_memsizes_valid = 1;
       bi.bi_bios_dev = saved_drive;
       bi.bi_basemem = mbi.mem_lower;
-      bi.bi_extmem = mbi.mem_upper;
+      bi.bi_extmem = extended_memory;
 
       if (mbi.flags & MB_INFO_AOUT_SYMS)
 	{
@@ -669,9 +793,16 @@ bsd_boot (kernel_t type, int bootdev, char *arg)
        */
 
       /* call entry point */
-      (*entry_addr) (clval, bootdev, 0,
-		     (mbi.syms.a.addr + 4
-		      + mbi.syms.a.tabsize + mbi.syms.a.strsize),
-		     mbi.mem_upper, mbi.mem_lower);
+      unsigned long end_mark;
+
+      if (mbi.flags & MB_INFO_AOUT_SYMS)
+	end_mark = (mbi.syms.a.addr + 4
+		    + mbi.syms.a.tabsize + mbi.syms.a.strsize);
+      else
+	/* FIXME: it should be mbi.syms.e.size.  */
+	end_mark = 0;
+      
+      (*entry_addr) (clval, bootdev, 0, end_mark,
+		     extended_memory, mbi.mem_lower);
     }
 }
