@@ -2,7 +2,7 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
  *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2000  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -19,9 +19,15 @@
  *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+#include <shared.h>
 
-#include "shared.h"
+#ifdef SUPPORT_SERIAL
+# include <serial.h>
+#endif
 
+#ifndef STAGE1_5
+int auto_fill = 1;
+#endif
 
 void
 print_error (void)
@@ -29,15 +35,12 @@ print_error (void)
   if (errnum > ERR_NONE && errnum < MAX_ERR_NUM)
 #ifndef STAGE1_5
     /* printf("\7\n %s\n", err_list[errnum]); */
-    printf ("\n %s\n", err_list[errnum]);
+    printf ("\nError %u: %s\n", errnum, err_list[errnum]);
 #else /* STAGE1_5 */
-    printf ("Error: %u\n", errnum);
+    printf ("Error %u\n", errnum);
 #endif /* STAGE1_5 */
-
-  errnum = ERR_NONE;
 }
 
-
 char *
 convert_to_ascii (char *buf, int c,...)
 {
@@ -45,7 +48,7 @@ convert_to_ascii (char *buf, int c,...)
   char *ptr = buf;
 
 #ifndef STAGE1_5
-  if (c == 'x')
+  if (c == 'x' || c == 'X' || c == 'b')
     mult = 16;
 
   if ((num & 0x80000000uL) && c == 'd')
@@ -69,9 +72,9 @@ convert_to_ascii (char *buf, int c,...)
     char *ptr2 = buf;
     while (ptr1 > ptr2)
       {
-	int c = *ptr1;
+	int tmp = *ptr1;
 	*ptr1 = *ptr2;
-	*ptr2 = c;
+	*ptr2 = tmp;
 	ptr1--;
 	ptr2++;
       }
@@ -80,13 +83,13 @@ convert_to_ascii (char *buf, int c,...)
   return ptr;
 }
 
-
 void
 grub_printf (const char *format,...)
 {
   int *dataptr = (int *) &format;
   char c, *ptr, str[16];
-
+  unsigned long mask = 0xFFFFFFFF;
+  
   dataptr++;
 
   while ((c = *(format++)) != 0)
@@ -97,11 +100,16 @@ grub_printf (const char *format,...)
 	switch (c = *(format++))
 	  {
 #ifndef STAGE1_5
+	  case 'b':
+	    mask = 0xFF;
+	    /* Fall down intentionally!  */
 	  case 'd':
 	  case 'x':
+	  case 'X':
 #endif
 	  case 'u':
-	    *convert_to_ascii (str, c, *((unsigned long *) dataptr++)) = 0;
+	    *convert_to_ascii (str, c, *((unsigned long *) dataptr++) & mask)
+	      = 0;
 
 	    ptr = str;
 
@@ -125,7 +133,6 @@ grub_printf (const char *format,...)
     }
 }
 
-
 #ifndef STAGE1_5
 int
 grub_sprintf (char *buffer, const char *format, ...)
@@ -170,11 +177,8 @@ grub_sprintf (char *buffer, const char *format, ...)
   *bp = 0;
   return bp - buffer;
 }
-#endif /* ! STAGE1_5 */
 
 
-#ifndef STAGE1_5
-
 void
 init_page (void)
 {
@@ -222,102 +226,305 @@ int
 get_cmdline (char *prompt, char *cmdline, int maxlen,
 	     int echo_char, int readline)
 {
-  int ystart, yend, xend, lpos, c;
+  /* This is a rather complicated function. So explain the concept.
+     
+     A command-line consists of ``section''s. A section is a part of the
+     line which may be displayed on the screen, but a section is never
+     displayed with another section simultaneously.
+
+     Each section is basically 77 or less characters, but the exception
+     is the first section, which is 78 or less characters, because the
+     starting point is special. See below.
+
+     The first section contains a prompt and a command-line (or the
+     first part of a command-line when it is too long to be fit in the
+     screen). So, in the first section, the number of command-line
+     characters displayed is 78 minus the length of the prompt (or
+     less). If the command-line has more characters, `>' is put at the
+     position 78 (zero-origin), to inform the user of the hidden
+     characters.
+
+     Other sections always have `<' at the first position, since there
+     is absolutely a section before each section. If there is a section
+     after another section, this section consists of 77 characters and
+     `>' at the last position. The last section has 77 or less
+     characters and doesn't have `>'.
+
+     Each section other than the last shares some characters with the
+     previous section. This region is called ``margin''. If the cursor
+     is put at the magin which is shared by the first section and the
+     second, the first section is displayed. Otherwise, a displayed
+     section is switched to another section, only if the cursor is put
+     outside that section.  */
+
+  /* XXX: These should be defined in shared.h, but I leave these here,
+     until this code is freezed.  */
+#define CMDLINE_WIDTH	78
+#define CMDLINE_MARGIN	10
+  
+  int xpos, lpos, c, section;
   /* The length of PROMPT.  */
-  int plen = 0;
+  int plen;
   /* The length of the command-line.  */
-  int llen = 0;
+  int llen;
   /* The index for the history.  */
   int history = -1;
   /* The working buffer for the command-line.  */
   char *buf = (char *) CMDLINE_BUF;
   /* The kill buffer.  */
-  char *kill = (char *) KILL_BUF;
+  char *kill_buf = (char *) KILL_BUF;
+  /* The original state of AUTO_FILL.  */
+  int saved_auto_fill = auto_fill;
+  
+  /* Nested function definitions for code simplicity.  */
 
-  /* nested function definition for code simplicity */
-  static void cl_print (char *str, int echo_char)
+  /* The forward declarations of nested functions are prefixed
+     with `auto'.  */
+  auto void cl_refresh (int full, int len);
+  auto void cl_backward (int count);
+  auto void cl_forward (int count);
+  auto void cl_insert (const char *str);
+  auto void cl_delete (int count);
+  auto void cl_init (void);
+  
+  /* Move the cursor backward.  */
+  void cl_backward (int count)
     {
-      while (*str != 0)
+      lpos -= count;
+      
+      /* If the cursor is in the first section, display the first section
+	 instead of the second.  */
+      if (section == 1 && plen + lpos < CMDLINE_WIDTH)
+	cl_refresh (1, 0);
+      else if (xpos - count < 1)
+	cl_refresh (1, 0);
+      else
 	{
-	  putchar (echo_char ? echo_char : *str);
-	  str++;
-	  if (++xend > 78)
+	  xpos -= count;
+
+	  if (terminal & TERMINAL_CONSOLE)
 	    {
-	      xend = 0;
-	      putchar (' ');
-	      if (yend == (getxy () & 0xff))
-		ystart--;
-	      else
-		yend++;
+	      int y = getxy () & 0xFF;
+	      
+	      gotoxy (xpos, y);
 	    }
+# ifdef SUPPORT_SERIAL
+	  else if (! (terminal & TERMINAL_DUMB) && (count > 4))
+	    {
+	      grub_printf ("\e[%dD", count);
+	    }
+	  else
+	    {
+	      int i;
+	      
+	      for (i = 0; i < count; i++)
+		grub_putchar ('\b');
+	    }
+# endif /* SUPPORT_SERIAL */
 	}
     }
 
-  /* nested function definition for code simplicity */
-  static void cl_setcpos (void)
+  /* Move the cursor forward.  */
+  void cl_forward (int count)
     {
-      yend = ((lpos + plen) / 79) + ystart;
-      xend = ((lpos + plen) % 79);
-      gotoxy (xend, yend);
+      lpos += count;
+
+      /* If the cursor goes outside, scroll the screen to the right.  */
+      if (xpos + count >= CMDLINE_WIDTH)
+	cl_refresh (1, 0);
+      else
+	{
+	  xpos += count;
+
+	  if (terminal & TERMINAL_CONSOLE)
+	    {
+	      int y = getxy () & 0xFF;
+	      
+	      gotoxy (xpos, y);
+	    }
+# ifdef SUPPORT_SERIAL
+	  else if (! (terminal & TERMINAL_DUMB) && (count > 4))
+	    {
+	      grub_printf ("\e[%dC", count);
+	    }
+	  else
+	    {
+	      int i;
+	      
+	      for (i = lpos - count; i < lpos; i++)
+		{
+		  if (! echo_char)
+		    grub_putchar (buf[i]);
+		  else
+		    grub_putchar (echo_char);
+		}
+	    }
+# endif /* SUPPORT_SERIAL */
+	}
     }
 
-  /* nested function definition for initial command-line printing */
-  static void cl_init ()
-    {
-      /* distinguish us from other lines and error messages! */
-      putchar ('\n');
-
-      /* print full line and set position here */
-      ystart = (getxy () & 0xff);
-      yend = ystart;
-      xend = 0;
-      cl_print (prompt, 0);
-      cl_print (buf, echo_char);
-      cl_setcpos ();
-    }
-
-  /* nested function definition for erasing to the end of the line */
-  static void cl_kill_to_end ()
+  /* Refresh the screen. If FULL is true, redraw the full line, otherwise,
+     only LEN characters from LPOS.  */
+  void cl_refresh (int full, int len)
     {
       int i;
-      buf[lpos] = 0;
-      for (i = lpos; i <= llen; i++)
+      int start;
+      int pos = xpos;
+      
+      if (full)
 	{
-	  if (i && ((i + plen) % 79) == 0)
-	    putchar (' ');
-	  putchar (' ');
+	  /* Recompute the section number.  */
+	  if (lpos + plen < CMDLINE_WIDTH)
+	    section = 0;
+	  else
+	    section = ((lpos + plen - CMDLINE_WIDTH)
+		       / (CMDLINE_WIDTH - 1 - CMDLINE_MARGIN) + 1);
+
+	  /* From the start to the end.  */
+	  len = CMDLINE_WIDTH;
+	  pos = 0;
+	  grub_putchar ('\r');
+
+	  /* If SECTION is the first section, print the prompt, otherwise,
+	     print `<'.  */
+	  if (section == 0)
+	    {
+	      grub_printf ("%s", prompt);
+	      len -= plen;
+	      pos += plen;
+	    }
+	  else
+	    {
+	      grub_putchar ('<');
+	      len--;
+	      pos++;
+	    }
 	}
-      llen = lpos;
-      cl_setcpos ();
+
+      /* Compute the index to start writing BUF and the resulting position
+	 on the screen.  */
+      if (section == 0)
+	{
+	  int offset = 0;
+	  
+	  if (! full)
+	    offset = xpos - plen;
+	  
+	  start = 0;
+	  xpos = lpos + plen;
+	  start += offset;
+	}
+      else
+	{
+	  int offset = 0;
+	  
+	  if (! full)
+	    offset = xpos - 1;
+	  
+	  start = ((section - 1) * (CMDLINE_WIDTH - 1 - CMDLINE_MARGIN)
+		   + CMDLINE_WIDTH - plen - CMDLINE_MARGIN);
+	  xpos = lpos + 1 - start;
+	  start += offset;
+	}
+
+      /* Print BUF. If ECHO_CHAR is not zero, put it instead.  */
+      for (i = start; i < start + len && i < llen; i++)
+	{
+	  if (! echo_char)
+	    grub_putchar (buf[i]);
+	  else
+	    grub_putchar (echo_char);
+
+	  pos++;
+	}
+
+      /* Fill up the rest of the line with spaces.  */
+      for (; i < start + len; i++)
+	{
+	  grub_putchar (' ');
+	  pos++;
+	}
+
+      /* If the cursor is at the last position, put `>' or a space,
+	 depending on if there are more characters in BUF.  */
+      if (pos == CMDLINE_WIDTH)
+	{
+	  if (start + len < llen)
+	    grub_putchar ('>');
+	  else
+	    grub_putchar (' ');
+	  
+	  pos++;
+	}
+
+      /* Back to XPOS.  */
+      if (terminal & TERMINAL_CONSOLE)
+	{
+	  int y = getxy () & 0xFF;
+	  
+	  gotoxy (xpos, y);
+	}
+# ifdef SUPPORT_SERIAL      
+      else if (! (terminal & TERMINAL_SERIAL) && (pos - xpos > 4))
+	{
+	  grub_printf ("\e[%dD", pos - xpos);
+	}
+      else
+	{
+	  for (i = 0; i < pos - xpos; i++)
+	    grub_putchar ('\b');
+	}
+# endif /* SUPPORT_SERIAL */
     }
 
-  static void cl_insert (const char *str)
+  /* Initialize the command-line.  */
+  void cl_init (void)
+    {
+      /* Distinguish us from other lines and error messages!  */
+      grub_putchar ('\n');
+
+      /* Print full line and set position here.  */
+      cl_refresh (1, 0);
+    }
+
+  /* Insert STR to BUF.  */
+  void cl_insert (const char *str)
     {
       int l = grub_strlen (str);
 
       if (llen + l < maxlen)
 	{
 	  if (lpos == llen)
-	    {
-	      grub_memmove (buf + lpos, str, l + 1);
-	      cl_setcpos ();
-	      cl_print (buf + lpos, echo_char);
-	      lpos += l;
-	      cl_setcpos ();
-	    }
+	    grub_memmove (buf + lpos, str, l + 1);
 	  else
 	    {
 	      grub_memmove (buf + lpos + l, buf + lpos, llen - lpos + 1);
 	      grub_memmove (buf + lpos, str, l);
-	      cl_setcpos ();
-	      cl_print (buf + lpos, echo_char);
-	      lpos += l;
-	      cl_setcpos ();
 	    }
+	  
 	  llen += l;
+	  lpos += l;
+	  if (xpos + l >= CMDLINE_WIDTH)
+	    cl_refresh (1, 0);
+	  else if (xpos + l + llen - lpos > CMDLINE_WIDTH)
+	    cl_refresh (0, CMDLINE_WIDTH - xpos);
+	  else
+	    cl_refresh (0, l + llen - lpos);
 	}
     }
 
+  /* Delete COUNT characters in BUF.  */
+  void cl_delete (int count)
+    {
+      grub_memmove (buf + lpos, buf + lpos + count, llen - count + 1);
+      llen -= count;
+      
+      if (xpos + llen + count - lpos > CMDLINE_WIDTH)
+	cl_refresh (0, CMDLINE_WIDTH - xpos);
+      else
+	cl_refresh (0, llen + count - lpos);
+    }
+
   plen = grub_strlen (prompt);
   llen = grub_strlen (cmdline);
 
@@ -333,40 +540,15 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
   lpos = llen;
   grub_strcpy (buf, cmdline);
 
+  /* Disable the auto fill mode.  */
+  auto_fill = 0;
+  
   cl_init ();
 
   while (ASCII_CHAR (c = getkey ()) != '\n' && ASCII_CHAR (c) != '\r')
     {
-      switch (c)
-	{
-	case KEY_LEFT:
-	  c = 2;
-	  break;
-	case KEY_RIGHT:
-	  c = 6;
-	  break;
-	case KEY_UP:
-	  c = 16;
-	  break;
-	case KEY_DOWN:
-	  c = 14;
-	  break;
-	case KEY_HOME:
-	  c = 1;
-	  break;
-	case KEY_END:
-	  c = 5;
-	  break;
-	case KEY_DC:
-	  c = 4;
-	  break;
-	case KEY_BACKSPACE:
-	  c = 8;
-	  break;
-	}
-
-      c = ASCII_CHAR (c);
-
+      c = translate_keycode (c);
+      
       /* If READLINE is non-zero, handle readline-like key bindings.  */
       if (readline)
 	{
@@ -405,11 +587,20 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 		for (i = lpos; i > 0 && buf[i - 1] != ' '; i--)
 		  ;
 
+		/* Invalidate the cache, because the user may exchange
+		   removable disks.  */
+		buf_drive = -1;
+
 		/* Copy this word to COMPLETION_BUFFER and do the
 		   completion.  */
 		grub_memmove (completion_buffer, buf + i, lpos - i);
 		completion_buffer[lpos - i] = 0;
+		/* Enable the auto fill mode temporarily.  */
+		auto_fill = 1;
 		ret = print_completions (is_filename, 1);
+		/* Disable the auto fill mode again.  */
+		auto_fill = 0;
+		errnum = ERR_NONE;
 
 		if (ret >= 0)
 		  {
@@ -418,15 +609,16 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 
 		    if (ret > 0)
 		      {
-			/* There is more than one candidates, so print
+			/* There are more than one candidates, so print
 			   the list.  */
 
-			/* Go to the part after the line here.  */
-			yend = ((llen + plen) / 79) + ystart;
 			grub_putchar ('\n');
-			gotoxy (0, getxy () & 0xff);
-
+			/* Enable the auto fill mode temporarily.  */
+			auto_fill = 1;
 			print_completions (is_filename, 0);
+			/* Disable the auto fill mode again.  */
+			auto_fill = 0;
+			errnum = ERR_NONE;
 		      }
 		  }
 
@@ -439,51 +631,43 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 	      }
 	      break;
 	    case 1:		/* C-a go to beginning of line */
-	      lpos = 0;
-	      cl_setcpos ();
+	      cl_backward (lpos);
 	      break;
 	    case 5:		/* C-e go to end of line */
-	      lpos = llen;
-	      cl_setcpos ();
+	      cl_forward (llen - lpos);
 	      break;
 	    case 6:		/* C-f forward one character */
 	      if (lpos < llen)
-		{
-		  lpos++;
-		  cl_setcpos ();
-		}
+		cl_forward (1);
 	      break;
 	    case 2:		/* C-b backward one character */
 	      if (lpos > 0)
-		{
-		  lpos--;
-		  cl_setcpos ();
-		}
+		cl_backward (1);
 	      break;
 	    case 21:		/* C-u kill to beginning of line */
 	      if (lpos == 0)
 		break;
-	      /* Copy the string being deleted to KILL.  */
-	      grub_memmove (kill, buf, lpos);
-	      kill[lpos] = 0;
-	      grub_memmove (buf, buf + lpos, llen - lpos + 1);
-	      lpos = llen - lpos;
-	      cl_setcpos ();
-	      cl_kill_to_end ();
-	      lpos = 0;
-	      cl_setcpos ();
-	      cl_print (buf, echo_char);
-	      cl_setcpos ();
+	      /* Copy the string being deleted to KILL_BUF.  */
+	      grub_memmove (kill_buf, buf, lpos);
+	      kill_buf[lpos] = 0;
+	      {
+		/* XXX: Not very clever.  */
+		
+		int count = lpos;
+		
+		cl_backward (lpos);
+		cl_delete (count);
+	      }
 	      break;
 	    case 11:		/* C-k kill to end of line */
 	      if (lpos == llen)
 		break;
-	      /* Copy the string being deleted to KILL.  */
-	      grub_memmove (kill, buf + lpos, llen - lpos + 1);
-	      cl_kill_to_end ();
+	      /* Copy the string being deleted to KILL_BUF.  */
+	      grub_memmove (kill_buf, buf + lpos, llen - lpos + 1);
+	      cl_delete (llen - lpos);
 	      break;
 	    case 25:		/* C-y yank the kill buffer */
-	      cl_insert (kill);
+	      cl_insert (kill_buf);
 	      break;
 	    case 16:		/* C-p fetch the previous command */
 	      {
@@ -504,14 +688,10 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 		    break;
 		  }
 
-		lpos = 0;
-		cl_setcpos ();
-		cl_kill_to_end ();
 		grub_strcpy (buf, p);
 		llen = grub_strlen (buf);
 		lpos = llen;
-		cl_print (buf, 0);
-		cl_setcpos ();
+		cl_refresh (1, 0);
 	      }
 	      break;
 	    case 14:		/* C-n fetch the next command */
@@ -531,14 +711,10 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 		if (! p)
 		  p = cmdline;
 
-		lpos = 0;
-		cl_setcpos ();
-		cl_kill_to_end ();
 		grub_strcpy (buf, p);
 		llen = grub_strlen (buf);
 		lpos = llen;
-		cl_print (buf, 0);
-		cl_setcpos ();
+		cl_refresh (1, 0);
 	      }
 	      break;
 	    }
@@ -554,28 +730,16 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 	case 4:		/* C-d delete character under cursor */
 	  if (lpos == llen)
 	    break;
-	  lpos++;
-	  /* fallthrough is on purpose! */
+	  cl_delete (1);
+	  break;
 	case 8:		/* C-h backspace */
-#ifdef GRUB_UTIL
+# ifdef GRUB_UTIL
 	case 127:	/* also backspace */
-#endif
+# endif
 	  if (lpos > 0)
 	    {
-	      int i;
-	      grub_memmove (buf + lpos - 1, buf + lpos, llen - lpos + 1);
-	      i = lpos;
-	      lpos = llen - 1;
-	      cl_setcpos ();
-	      putchar (' ');
-	      lpos = i - 1;	/* restore lpos and decrement */
-	      llen--;
-	      cl_setcpos ();
-	      if (lpos != llen)
-		{
-		  cl_print (buf + lpos, echo_char);
-		  cl_setcpos ();
-		}
+	      cl_backward (1);
+	      cl_delete (1);
 	    }
 	  break;
 	default:		/* insert printable character into line */
@@ -590,11 +754,7 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
 	}
     }
 
-  /* Move the cursor to the end of the command.  */
-  lpos = llen;
-  cl_setcpos ();
   grub_putchar ('\n');
-  gotoxy (0, getxy () & 0xff);
 
   /* If ECHO_CHAR is NUL, remove the leading spaces.  */
   lpos = 0;
@@ -610,11 +770,126 @@ get_cmdline (char *prompt, char *cmdline, int maxlen,
   if (readline && lpos < llen)
     add_history (cmdline, 0);
 
+  /* Restore the auto fill mode.  */
+  auto_fill = saved_auto_fill;
+  
   return 0;
 }
 
-#endif /* STAGE1_5 */
+/* Translate a special key to a common ascii code.  */
+int
+translate_keycode (int c)
+{
+# ifdef SUPPORT_SERIAL
+  if (terminal & TERMINAL_SERIAL)
+    {
+      /* In a serial terminal, things are complicated, because several
+	 key codes start from the character ESC, while we want to accept
+	 ESC itself.  */
+      if (c == '\e')
+	{
+	  int start;
 
+	  /* Get current time.  */
+	  start = currticks ();
+
+	  while (checkkey () == -1)
+	    {
+	      /* Wait for a next character, at least for 0.1 sec
+		 (18.2 ticks/sec).  */
+	      int now;
+	      
+	      now = currticks ();
+	      if (now - start >= 2)
+		return c;
+	    }
+
+	  c = getkey ();
+	  if (c == '[')
+	    {
+	      int c1, c2;
+
+	      /* To filter illegal states.  */
+	      c = 0;
+	      c1 = getkey ();
+	      switch (c1)
+		{
+		case 'A':	/* KEY_UP */
+		  c = 16;
+		  break;
+		case 'B':	/* KEY_DOWN */
+		  c = 14;
+		  break;
+		case 'C':	/* KEY_RIGHT */
+		  c = 6;
+		  break;
+		case 'D':	/* KEY_LEFT */
+		  c = 2;
+		  break;
+		case 'F':	/* End */
+		  c = 5;
+		  break;
+		case 'H':	/* Home */
+		  c = 1;
+		  break;
+		case '1':
+		  c2 = getkey ();
+		  if (c2 == '~')
+		    {
+		      /* One of control keys (pos1,....).  */
+		      c = 1;
+		    }
+		  break;
+		case '3':
+		  c2 = getkey ();
+		  if (c2 == '~')
+		    {
+		      /* One of control keys (del,....).  */
+		      c = 4;
+		    }
+		  break;
+		case '4':	/* Del */
+		  c = 4;
+		  break;
+		}
+	    }
+	}
+    }
+  else
+# endif /* SUPPORT_SERIAL */
+    {
+      switch (c)
+	{
+	case KEY_LEFT:
+	  c = 2;
+	  break;
+	case KEY_RIGHT:
+	  c = 6;
+	  break;
+	case KEY_UP:
+	  c = 16;
+	  break;
+	case KEY_DOWN:
+	  c = 14;
+	  break;
+	case KEY_HOME:
+	  c = 1;
+	  break;
+	case KEY_END:
+	  c = 5;
+	  break;
+	case KEY_DC:
+	  c = 4;
+	  break;
+	case KEY_BACKSPACE:
+	  c = 8;
+	  break;
+	}
+    }
+  
+  return ASCII_CHAR (c);
+}
+#endif /* STAGE1_5 */
 
 int
 safe_parse_maxint (char **str_ptr, int *myint_ptr)
@@ -670,7 +945,6 @@ safe_parse_maxint (char **str_ptr, int *myint_ptr)
   return 1;
 }
 
-
 int
 grub_tolower (int c)
 {
@@ -680,11 +954,10 @@ grub_tolower (int c)
   return c;
 }
 
-
 int
 grub_isspace (int c)
 {
-  if (c == ' ' || c == '\t' || c == '\n')
+  if (c == ' ' || c == '\t' || c == '\r' || c == '\n')
     return 1;
 
   return 0;
@@ -707,9 +980,7 @@ grub_memcmp (const char *s1, const char *s2, int n)
 
   return 0;
 }
-#endif /* ! STAGE1_5 */
 
-#if 0
 int
 grub_strncat (char *s1, const char *s2, int n)
 {
@@ -728,9 +999,7 @@ grub_strncat (char *s1, const char *s2, int n)
 
   return 1;
 }
-#endif
 
-#ifndef STAGE1_5
 int
 grub_strcmp (const char *s1, const char *s2)
 {
@@ -746,6 +1015,234 @@ grub_strcmp (const char *s1, const char *s2)
 
   return 0;
 }
+
+/* Wait for a keypress and return its code.  */
+int
+getkey (void)
+{
+  int c = -1;
+  
+  if (terminal & TERMINAL_CONSOLE)
+    c = console_getkey ();
+#ifdef SUPPORT_SERIAL
+  else if (terminal & TERMINAL_SERIAL)
+    c = serial_getkey ();
+#endif /* SUPPORT_SERIAL */
+
+  return c;
+}
+
+/* Check if a key code is available.  */
+int
+checkkey (void)
+{
+  int c = -1;
+
+  if (terminal & TERMINAL_CONSOLE)
+    c = console_checkkey ();
+
+#ifdef SUPPORT_SERIAL
+  if (terminal & TERMINAL_SERIAL)
+    c = serial_checkkey ();
+#endif /* SUPPORT_SERIAL */
+
+  return c;
+}
+
+#endif /* ! STAGE1_5 */
+
+/* Display an ASCII character.  */
+void
+grub_putchar (int c)
+{
+#ifndef STAGE1_5
+  static int col = 0;
+#endif
+  
+  if (c == '\n')
+    grub_putchar ('\r');
+  
+#ifdef STAGE1_5
+  
+  /* In Stage 1.5, only the normal console is supported.  */
+  console_putchar (c);
+  
+#else /* ! STAGE1_5 */
+
+  /* Track the cursor by software here.  */
+  /* XXX: This doesn't handle horizontal or vertical tabs.  */
+  if (c == '\r')
+    col = 0;
+  else if (c == '\b')
+    {
+      if (col > 0)
+	col--;
+    }
+  else if (c >= ' ' && c <= '~')
+    {
+      /* Fold a line only if AUTO_FILL is true.  */
+      if (auto_fill && col >= 79)
+	grub_putchar ('\n');
+
+      col++;
+    }
+
+  if (terminal & TERMINAL_CONSOLE)
+    console_putchar (c);
+
+# ifdef SUPPORT_SERIAL
+  if (terminal & TERMINAL_SERIAL)
+    serial_putchar (c);
+# endif /* SUPPORT_SERIAL */
+  
+#endif /* ! STAGE1_5 */
+}
+
+#ifndef STAGE1_5
+void
+gotoxy (int x, int y)
+{
+  if (terminal & TERMINAL_CONSOLE)
+    console_gotoxy (x, y);
+#ifdef SUPPORT_SERIAL
+  else if (terminal & TERMINAL_SERIAL)
+    serial_gotoxy (x, y);
+#endif
+}
+
+#ifdef SUPPORT_SERIAL
+/* The serial part of gotoxy.  */
+void
+serial_gotoxy (int x, int y)
+{
+  grub_printf ("\e[%d;%dH", y + 1, x + 1);
+}
+#endif /* SUPPORT_SERIAL */
+
+int
+getxy (void)
+{
+  int ret = 0;
+  
+  if (terminal & TERMINAL_CONSOLE)
+    ret = console_getxy ();
+#ifdef SUPPORT_SERIAL
+  else if (terminal & TERMINAL_SERIAL)
+    ret = serial_getxy ();
+#endif
+  
+  return ret;
+}
+
+#ifdef SUPPORT_SERIAL
+/* The serial part of getxy.  */
+int
+serial_getxy (void)
+{
+  int x, y;
+  int start, now;
+  char buf[32];	/* XXX */
+  int i;
+  int c;
+  char *p;
+  
+  /* Drain the input buffer.  */
+  while (serial_checkkey () != -1)
+    serial_getkey ();
+
+  /* CPR.  */
+  grub_printf ("\e[6n");
+
+  /* Get current time.  */
+  while ((start = getrtsecs ()) == 0xFF)
+    ;
+
+ again:
+  i = 0;
+  c = 0;
+  do
+    {
+      if (serial_checkkey () != -1)
+	{
+	  c = serial_getkey ();
+	  if (i == 1 && c != '[')
+	    i = 0;
+	  
+	  if (i == 0 && c != '\e')
+	    continue;
+
+	  if (i != 0 && c == '\e')
+	    i = 0;
+
+	  buf[i] = c;
+	  i++;
+	}
+      else
+	{
+	  /* Get current time.  */
+	  while ((now = getrtsecs ()) == 0xFF)
+	    ;
+
+	  /* FIXME: Remove this magic number.  */
+	  if (now - start > 10)
+	    {
+	      /* Something is quite wrong.  */
+	      return 0;
+	    }
+	}
+    }
+  while (c != 'R' && i < sizeof (buf));
+  
+  if (c != 'R')
+    goto again;
+
+  p = buf + 2;
+  if (! safe_parse_maxint (&p, &y))
+    {
+      errnum = 0;
+      goto again;
+    }
+  
+  if (*p != ';')
+    goto again;
+
+  p++;
+  if (! safe_parse_maxint (&p, &x))
+    {
+      errnum = 0;
+      goto again;
+    }
+
+  if (*p != 'R')
+    goto again;
+
+  return ((x - 1) << 8) | (y - 1);
+}
+#endif /* SUPPORT_SERIAL */
+
+void
+cls (void)
+{
+  if (terminal & TERMINAL_CONSOLE)
+    console_cls ();
+#ifdef SUPPORT_SERIAL
+  else if (terminal & TERMINAL_SERIAL)
+    serial_cls ();
+#endif
+}
+
+#ifdef SUPPORT_SERIAL
+/* The serial part of cls.  */
+void
+serial_cls (void)
+{
+  /* If the terminal is dumb, there is no way to clean the terminal.  */
+  if (terminal & TERMINAL_DUMB)
+    grub_putchar ('\n');
+  else
+    grub_printf ("\e[H\e[J");
+}
+#endif /* SUPPORT_SERIAL */
 #endif /* ! STAGE1_5 */
 
 int
@@ -767,8 +1264,21 @@ substring (char *s1, char *s2)
   return 1;
 }
 
-
 #ifndef STAGE1_5
+/* Terminate the string STR with NUL.  */
+int
+nul_terminate (char *str)
+{
+  int ch;
+  
+  while (*str && ! grub_isspace (*str))
+    str++;
+
+  ch = *str;
+  *str = 0;
+  return ch;
+}
+
 char *
 grub_strstr (const char *s1, const char *s2)
 {
@@ -800,7 +1310,6 @@ grub_strlen (const char *str)
 }
 #endif /* ! STAGE1_5 */
 
-
 int
 memcheck (int addr, int len)
 {
@@ -833,17 +1342,16 @@ memcheck (int addr, int len)
 
   if ((addr < RAW_ADDR (0x1000))
       || (addr < RAW_ADDR (0x100000)
-	  && RAW_ADDR (mbi.mem_lower * 1024) <= (addr + len))
+	  && RAW_ADDR (mbi.mem_lower * 1024) < (addr + len))
       || (addr >= RAW_ADDR (0x100000)
-	  && RAW_ADDR (mbi.mem_upper * 1024) <= ((addr - 0x100000) + len)))
+	  && RAW_ADDR (mbi.mem_upper * 1024) < ((addr - 0x100000) + len)))
     errnum = ERR_WONT_FIT;
 
   return ! errnum;
 }
 
-
-char *
-grub_memmove (char *to, const char *from, int len)
+void *
+grub_memmove (void *to, const void *from, int len)
 {
    if (memcheck ((int) to, len))
      {
@@ -869,8 +1377,8 @@ grub_memmove (char *to, const char *from, int len)
 			 "cld"
 			 : "=&c" (d0), "=&S" (d1), "=&D" (d2)
 			 : "0" (len),
-			 "1" (len - 1 + from),
-			 "2" (len - 1 + to)
+			 "1" (len - 1 + (const char *) from),
+			 "2" (len - 1 + (char *) to)
 			 : "memory");
 	 }
      }
@@ -878,7 +1386,6 @@ grub_memmove (char *to, const char *from, int len)
    return errnum ? NULL : to;
 }
 
-
 #ifndef STAGE1_5
 void *
 grub_memset (void *start, int c, int len)
