@@ -2,7 +2,7 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
  *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2000  Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -26,8 +26,8 @@
 
 #ifndef STAGE1_5
 /* instrumentation variables */
-void (*disk_read_hook) (int) = NULL;
-void (*disk_read_func) (int) = NULL;
+void (*disk_read_hook) (int, int, int) = NULL;
+void (*disk_read_func) (int, int, int) = NULL;
 
 int print_possibilities;
 
@@ -42,23 +42,26 @@ struct fsys_entry fsys_table[NUM_FSYS + 1] =
 {
   /* TFTP should come first because others don't handle net device.  */
 # ifdef FSYS_TFTP
-  {"tftp", tftp_mount, tftp_read, tftp_dir},
+  {"tftp", tftp_mount, tftp_read, tftp_dir, tftp_close, 0},
 # endif
 # ifdef FSYS_FAT
-  {"fat", fat_mount, 0, fat_dir},
+  {"fat", fat_mount, fat_read, fat_dir, 0, 0},
 # endif
 # ifdef FSYS_EXT2FS
-  {"ext2fs", ext2fs_mount, ext2fs_read, ext2fs_dir},
+  {"ext2fs", ext2fs_mount, ext2fs_read, ext2fs_dir, 0, 0},
 # endif
 # ifdef FSYS_MINIX
-  {"minix", minix_mount, minix_read, minix_dir},
+  {"minix", minix_mount, minix_read, minix_dir, 0, 0},
+# endif
+# ifdef FSYS_REISERFS
+  {"reiserfs", reiserfs_mount, reiserfs_read, reiserfs_dir, 0, reiserfs_embed},
 # endif
   /* XX FFS should come last as it's superblock is commonly crossing tracks
      on floppies from track 1 to 2, while others only use 1.  */
 # ifdef FSYS_FFS
-  {"ffs", ffs_mount, ffs_read, ffs_dir},
+  {"ffs", ffs_mount, ffs_read, ffs_dir, 0, ffs_embed},
 # endif
-  {0, 0, 0, 0}
+  {0, 0, 0, 0, 0, 0}
 };
 
 
@@ -67,6 +70,13 @@ struct fsys_entry fsys_table[NUM_FSYS + 1] =
 unsigned long current_drive = 0xFF;
 unsigned long current_partition;
 
+#ifndef STAGE1_5
+/* The register ESI should contain the address of the partition to be
+   used for loading a chain-loader when chain-loading the loader.  */
+unsigned long boot_part_addr;
+unsigned long boot_part_offset;
+#endif
+
 /*
  *  Global variables describing details of the filesystem
  */
@@ -78,7 +88,7 @@ int bsd_evil_hack;
 /* filesystem type */
 int fsys_type = NUM_FSYS;
 #ifndef NO_BLOCK_FILES
-int block_file = 0;
+static int block_file = 0;
 #endif /* NO_BLOCK_FILES */
 
 /* these are the translated numbers for the open partition */
@@ -100,7 +110,7 @@ int filemax;
 int
 rawread (int drive, int sector, int byte_offset, int byte_len, char *buf)
 {
-  int slen = (byte_offset + byte_len + SECTOR_SIZE - 1) / SECTOR_SIZE;
+  int slen = (byte_offset + byte_len + SECTOR_SIZE - 1) >> SECTOR_BITS;
 
   if (byte_len <= 0)
     return 1;
@@ -165,30 +175,39 @@ rawread (int drive, int sector, int byte_offset, int byte_len, char *buf)
 				   sector, slen, BUFFERSEG))
 		    errnum = ERR_READ;
 
-		  bufaddr = BUFFERSEG + byte_offset;
+		  bufaddr = BUFFERADDR + byte_offset;
 		}
 	    }
 	  else
 	    buf_track = track;
 	}
 
+      if (size > ((num_sect * SECTOR_SIZE) - byte_offset))
+	size = (num_sect * SECTOR_SIZE) - byte_offset;
+
 #ifndef STAGE1_5
       /*
        *  Instrumentation to tell which sectors were read and used.
        */
-      if (disk_read_hook && disk_read_func)
+      if (disk_read_func)
 	{
-	  int sector_end = sector + ((num_sect < slen) ? num_sect : slen);
 	  int sector_num = sector;
-
-	  while (sector_num < sector_end)
-	    (*disk_read_func) (sector_num++);
+	  int length = SECTOR_SIZE - byte_offset;
+	  if (length > size)
+	    length = size;
+	  (*disk_read_func) (sector_num++, byte_offset, length);
+	  length = size - length;
+	  if (length > 0)
+	    {
+	      while (length > SECTOR_SIZE)
+		{
+		  (*disk_read_func) (sector_num++, 0, SECTOR_SIZE);
+		  length -= SECTOR_SIZE;
+		}
+	      (*disk_read_func) (sector_num, 0, length);
+	    }
 	}
 #endif /* STAGE1_5 */
-
-      if (size > ((num_sect * SECTOR_SIZE) - byte_offset))
-	size = (num_sect * SECTOR_SIZE) - byte_offset;
-
       memmove (buf, (char *) bufaddr, size);
 
       buf += size;
@@ -209,7 +228,8 @@ devread (int sector, int byte_offset, int byte_len, char *buf)
    *  Check partition boundaries
    */
   if (sector < 0
-      || (sector + ((byte_offset + byte_len - 1) / SECTOR_SIZE)) >= part_length)
+      || ((sector + ((byte_offset + byte_len - 1) >> SECTOR_BITS))
+	  >= part_length))
     {
       errnum = ERR_OUTSIDE_PART;
       return 0;
@@ -218,11 +238,8 @@ devread (int sector, int byte_offset, int byte_len, char *buf)
   /*
    *  Get the read to the beginning of a partition.
    */
-  while (byte_offset >= SECTOR_SIZE)
-    {
-      byte_offset -= SECTOR_SIZE;
-      sector++;
-    }
+  sector += byte_offset >> SECTOR_BITS;
+  byte_offset &= SECTOR_SIZE - 1;
 
 #if !defined(STAGE1_5)
   if (disk_read_hook && debug)
@@ -247,7 +264,7 @@ static int
 sane_partition (void)
 {
   /* network drive */
-  if (current_drive == 0x20)
+  if (current_drive == NETWORK_DRIVE)
     return 1;
   
   if (!(current_partition & 0xFF000000uL)
@@ -285,33 +302,53 @@ attempt_mount (void)
 
 
 #ifndef STAGE1_5
+/* Turn on the active flag for the partition SAVED_PARATITION in the
+   drive SAVED_DRIVE. If an error occurs, return zero, otherwise return
+   non-zero.  */
 int
 make_saved_active (void)
 {
   if (saved_drive & 0x80)
     {
+      /* Hard disk */
       int part = saved_partition >> 16;
 
+      /* If the partition is not a primary partition, the active flag is
+	 meaningless. (XXX: Really?)  */
       if (part > 3)
 	{
 	  errnum = ERR_NO_PART;
 	  return 0;
 	}
 
-      if (!rawread (saved_drive, 0, 0, SECTOR_SIZE, (char *) SCRATCHADDR))
+      /* Read the MBR in the scratch space.  */
+      if (! rawread (saved_drive, 0, 0, SECTOR_SIZE, (char *) SCRATCHADDR))
 	return 0;
 
+      /* If the partition is an extended partition, setting the active
+	 flag violates the specification by IBM.  */
+      if (IS_PC_SLICE_TYPE_EXTENDED (PC_SLICE_TYPE (SCRATCHADDR, part)))
+	{
+	  errnum = ERR_DEV_VALUES;
+	  return 0;
+	}
+
+      /* Check if the active flag is disabled.  */
       if (PC_SLICE_FLAG (SCRATCHADDR, part) != PC_SLICE_FLAG_BOOTABLE)
 	{
 	  int i;
 
+	  /* Clear all the active flags in this table.  */
 	  for (i = 0; i < 4; i++)
 	    PC_SLICE_FLAG (SCRATCHADDR, i) = 0;
 
+	  /* Set the flag.  */
 	  PC_SLICE_FLAG (SCRATCHADDR, part) = PC_SLICE_FLAG_BOOTABLE;
 
+	  /* Clear the cache.  */
 	  buf_track = -1;
 
+	  /* Write back the MBR.  */
 	  if (biosdisk (BIOSDISK_WRITE, saved_drive, &buf_geom,
 			0, 1, SCRATCHSEG))
 	    {
@@ -320,16 +357,24 @@ make_saved_active (void)
 	    }
 	}
     }
+  else
+    {
+      /* If the drive is not a hard disk drive, you shouldn't call this
+	 function. (XXX: Should I just ignore this error?)  */
+      errnum = ERR_DEV_VALUES;
+      return 0;
+    }
 
   return 1;
 }
 
+/* Hide/Unhide CURRENT_PARTITION.  */
 int
 set_partition_hidden_flag (int hidden)
 {
-  if (saved_drive)
+  if (current_drive & 0x80)
     {
-      int part = saved_partition >> 16;
+      int part = current_partition >> 16;
 
       if (part > 3)
 	{
@@ -337,7 +382,7 @@ set_partition_hidden_flag (int hidden)
           return 0;
         }
 
-      if (! rawread (saved_drive, 0, 0, SECTOR_SIZE, (char *) SCRATCHADDR))
+      if (! rawread (current_drive, 0, 0, SECTOR_SIZE, (char *) SCRATCHADDR))
         return 0;
 
       if (hidden)
@@ -346,7 +391,7 @@ set_partition_hidden_flag (int hidden)
 	PC_SLICE_TYPE (SCRATCHADDR, part) &= ~PC_SLICE_TYPE_HIDDEN_FLAG;
       
       buf_track = -1;
-      if (biosdisk (BIOSDISK_WRITE, saved_drive, &buf_geom,
+      if (biosdisk (BIOSDISK_WRITE, current_drive, &buf_geom,
 		    0, 1, SCRATCHSEG))
 	{
 	  errnum = ERR_WRITE;
@@ -371,287 +416,362 @@ check_and_print_mount (void)
 #endif /* STAGE1_5 */
 
 
-static int
-check_BSD_parts (int flags)
+/* Get the information on next partition on the drive DRIVE.
+   The caller must not modify the contents of the arguments when
+   iterating this function. The partition representation in GRUB will
+   be stored in *PARTITION. Likewise, the partition type in *TYPE, the
+   start sector in *START, the length in *LEN, the offset of the
+   partition table in *OFFSET, the entry number in the table in *ENTRY,
+   the offset of the extended partition in *EXT_OFFSET.
+   BUF is used to store a MBR, the boot sector of a partition, or
+   a BSD label sector, and it must be at least 512 bytes length.
+   When calling this function first, *PARTITION must be initialized to
+   0xFFFFFF. The return value is zero if fails, otherwise non-zero.  */
+int
+next_partition (unsigned long drive, unsigned long dest,
+		unsigned long *partition, int *type,
+		unsigned long *start, unsigned long *len,
+		unsigned long *offset, int *entry,
+		unsigned long *ext_offset, char *buf)
 {
-  char label_buf[SECTOR_SIZE];
-  int part_no, got_part = 0;
+  /* Forward declarations.  */
+  auto int next_bsd_partition (void);
+  auto int next_pc_slice (void);
 
-  if (part_length < (BSD_LABEL_SECTOR + 1))
+  /* Get next BSD partition in current PC slice.  */
+  int next_bsd_partition (void)
     {
-      errnum = ERR_BAD_PART_TABLE;
-      return 0;
-    }
+      int i;
+      int bsd_part_no = (*partition & 0xFF00) >> 8;
 
-  if (!rawread (current_drive, part_start + BSD_LABEL_SECTOR,
-		0, SECTOR_SIZE, label_buf))
-    return 0;
-
-  if (BSD_LABEL_CHECK_MAG (label_buf))
-    {
-      for (part_no = 0; part_no < BSD_LABEL_NPARTS (label_buf); part_no++)
+      /* If this is the first time...  */
+      if (bsd_part_no == 0xFF)
 	{
-	  if (BSD_PART_TYPE (label_buf, part_no))
+	  /* Check if the BSD label is within current PC slice.  */
+	  if (*len < BSD_LABEL_SECTOR + 1)
 	    {
-	      /* FIXME: should do BAD144 sector remapping setup here */
+	      errnum = ERR_BAD_PART_TABLE;
+	      return 0;
+	    }
 
-	      /* XXX We cannot determine which variant of BSD owns
-		 this slice, so set it to FreeBSD paritition type.
-		 That should work fine for now.  */
-	      current_slice = ((BSD_PART_TYPE (label_buf, part_no) << 8)
-			       | PC_SLICE_TYPE_FREEBSD);
-	      part_start = BSD_PART_START (label_buf, part_no);
-	      part_length = BSD_PART_LENGTH (label_buf, part_no);
+	  /* Read the BSD label.  */
+	  if (! rawread (drive, *start + BSD_LABEL_SECTOR,
+			 0, SECTOR_SIZE, buf))
+	    return 0;
+
+	  /* Check if it is valid.  */
+	  if (! BSD_LABEL_CHECK_MAG (buf))
+	    {
+	      errnum = ERR_BAD_PART_TABLE;
+	      return 0;
+	    }
+	  
+	  bsd_part_no = -1;
+	}
+
+      /* Search next valid BSD partition.  */
+      for (i = bsd_part_no + 1; i < BSD_LABEL_NPARTS (buf); i++)
+	{
+	  if (BSD_PART_TYPE (buf, i))
+	    {
+	      /* Note that *TYPE and *PARTITION were set
+		 for current PC slice.  */
+	      *type = (BSD_PART_TYPE (buf, i) << 8) | (*type & 0xFF);
+	      *start = BSD_PART_START (buf, i);
+	      *len = BSD_PART_LENGTH (buf, i);
+	      *partition = (*partition & 0xFF00FF) | (i << 8);
 
 #ifndef STAGE1_5
-	      if (flags)
-		{
-		  if (!got_part)
-		    {
-		      if (! do_completion)
-			printf ("[BSD sub-partitions immediately follow]\n");
-		      got_part = 1;
-		    }
-
-		  if (! do_completion)
-		    printf ("     BSD Partition num: \'%c\', ", part_no + 'a');
-		  else
-		    {
-		      char str[16];
-
-		      grub_sprintf (str, "%d,%c)",
-				    (current_partition >> 16) & 0xFF,
-				    part_no + 'a');
-		      print_a_completion (str);
-		    }
-		  
-		  check_and_print_mount ();
-		}
-	      else
-#endif /* STAGE1_5 */
-	      if (part_no == ((current_partition >> 8) & 0xFF))
-		break;
+	      /* XXX */
+	      if ((drive & 0x80) && BSD_LABEL_DTYPE (buf) == DTYPE_SCSI)
+		bsd_evil_hack = 4;
+#endif /* ! STAGE1_5 */
+	      
+	      return 1;
 	    }
 	}
 
-      if (part_no >= BSD_LABEL_NPARTS (label_buf) && !got_part)
+      errnum = ERR_NO_PART;
+      return 0;
+    }
+
+  /* Get next PC slice. Be careful of that this function may return
+     an empty PC slice (i.e. a partition whose type is zero) as well.  */
+  int next_pc_slice (void)
+    {
+      int pc_slice_no = (*partition & 0xFF0000) >> 16;
+
+      /* If this is the first time...  */
+      if (pc_slice_no == 0xFF)
 	{
-	  errnum = ERR_NO_PART;
+	  *offset = 0;
+	  *ext_offset = 0;
+	  *entry = -1;
+	  pc_slice_no = -1;
+	}
+
+      /* Read the MBR or the boot sector of the extended partition.  */
+      if (! rawread (drive, *offset, 0, SECTOR_SIZE, buf))
+	return 0;
+
+      /* Check if it is valid.  */
+      if (! PC_MBR_CHECK_SIG (buf))
+	{
+	  errnum = ERR_BAD_PART_TABLE;
 	  return 0;
 	}
 
-      if ((current_drive & 0x80)
-	  && BSD_LABEL_DTYPE (label_buf) == DTYPE_SCSI)
-	bsd_evil_hack = 4;
+      /* Increase the entry number.  */
+      (*entry)++;
 
+      /* If this is out of current partition table...  */
+      if (*entry == PC_SLICE_MAX)
+	{
+	  int i;
+
+	  /* Search the first extended partition in current table.  */
+	  for (i = 0; i < PC_SLICE_MAX; i++)
+	    {
+	      if (IS_PC_SLICE_TYPE_EXTENDED (PC_SLICE_TYPE (buf, i)))
+		{
+		  /* Found. Set the new offset and the entry number,
+		     and restart this function.  */
+		  *offset = *ext_offset + PC_SLICE_START (buf, i);
+		  if (! *ext_offset)
+		    *ext_offset = *offset;
+		  *entry = -1;
+		  return next_pc_slice ();
+		}
+	    }
+
+	  errnum = ERR_NO_PART;
+	  return 0;
+	}
+      
+      *type = PC_SLICE_TYPE (buf, *entry);
+      *start = *offset + PC_SLICE_START (buf, *entry);
+      *len = PC_SLICE_LENGTH (buf, *entry);
+
+      /* The calculation of a PC slice number is complicated, because of
+	 the rather odd definition of extended partitions. Even worse,
+	 there is no guarantee that this is consistent with every
+	 operating systems. Uggh.  */
+      if (pc_slice_no < PC_SLICE_MAX
+	  || (! IS_PC_SLICE_TYPE_EXTENDED (*type)
+	      && *type != PC_SLICE_TYPE_NONE))
+	pc_slice_no++;
+      
+      *partition = (pc_slice_no << 16) | 0xFFFF;
       return 1;
     }
 
-  errnum = ERR_BAD_PART_TABLE;
-  return 0;
-}
-
-
+  /* Start the body of this function.  */
+  
 #ifndef STAGE1_5
-static char cur_part_desc[16];
+  if (current_drive == NETWORK_DRIVE)
+    return 0;
 #endif
 
+  /* If previous partition is a BSD partition or a PC slice which
+     contains BSD partitions...  */
+  if ((*partition != 0xFFFFFF && IS_PC_SLICE_TYPE_BSD (*type))
+      || ! (drive & 0x80))
+    {
+      if (*type == PC_SLICE_TYPE_NONE)
+	*type = PC_SLICE_TYPE_FREEBSD;
+      
+      /* Get next BSD partition, if any.  */
+      if (next_bsd_partition ())
+	return 1;
+
+      /* If the destination partition is a BSD partition and current
+	 BSD partition has any error, abort the operation.  */
+      if ((dest & 0xFF00) != 0xFF00
+	  && ((dest & 0xFF0000) == 0xFF0000
+	      || (dest & 0xFF0000) == (*partition & 0xFF0000)))
+	return 0;
+
+      /* Ignore the error.  */
+      errnum = ERR_NONE;
+    }
+	  
+  return next_pc_slice ();
+}
+
+#ifndef STAGE1_5
+static unsigned long cur_part_offset;
+static unsigned long cur_part_addr;
+#endif
+
+/* Open a partition.  */
 int
 real_open_partition (int flags)
 {
-  char mbr_buf[SECTOR_SIZE];
-  int i, part_no, slice_no, ext = 0;
+  unsigned long dest_partition = current_partition;
+  unsigned long part_offset;
+  unsigned long ext_offset;
+  int entry;
+  char buf[SECTOR_SIZE];
+  int bsd_part, pc_slice;
 
+  /* For simplicity.  */
+  auto int next (void);
+  int next (void)
+    {
+      int ret = next_partition (current_drive, dest_partition,
+				&current_partition, &current_slice,
+				&part_start, &part_length,
+				&part_offset, &entry, &ext_offset, buf);
+      bsd_part = (current_partition >> 8) & 0xFF;
+      pc_slice = current_partition >> 16;
+      return ret;
+    }
+  
 #ifndef STAGE1_5
   /* network drive */
-  if (current_drive == 0x20)
+  if (current_drive == NETWORK_DRIVE)
     return 1;
   
   if (! sane_partition ())
     return 0;
 #endif
 
-  /*
-   *  The "rawread" is probably unnecessary here, but it is good to
-   *  know it works.
-   */
-  if (! rawread (current_drive, 0, 0, SECTOR_SIZE, mbr_buf))
-    return 0;
-
   bsd_evil_hack = 0;
   current_slice = 0;
   part_start = 0;
+
+  /* Make sure that buf_geom is valid. */
+  if (buf_drive != current_drive)
+    {
+      if (get_diskinfo (current_drive, &buf_geom))
+	{
+	  errnum = ERR_NO_DISK;
+	  return 0;
+	}
+      buf_drive = current_drive;
+      buf_track = -1;
+    }
   part_length = buf_geom.total_sectors;
 
-  if (current_drive & 0x80)
+  /* If this is the whole disk, return here.  */
+  if (! flags && current_partition == 0xFFFFFF)
+    return 1;
+
+  if (flags)
+    dest_partition = 0xFFFFFF;
+
+  /* Initialize CURRENT_PARTITION for next_partition.  */
+  current_partition = 0xFFFFFF;
+  
+  while (next ())
     {
-      /*
-       *  We're looking at a hard disk
-       */
-
-      int ext_offset = 0, part_offset = 0;
-      part_no = (current_partition >> 16);
-      slice_no = 0;
-
-      /* if this is the whole disk, return here */
-      if (!flags && current_partition == 0xFFFFFFuL)
-	return 1;
-
-      /*
-       *  Load the current MBR-style PC partition table (4 entries)
-       */
-      while (slice_no < 255 && ext >= 0
-	     && (part_no == 0xFF || slice_no <= part_no)
-	     && rawread (current_drive, part_offset,
-			 0, SECTOR_SIZE, mbr_buf))
-	{
-	  /*
-	   *  If the table isn't valid, we can't continue
-	   */
-	  if (!PC_MBR_CHECK_SIG (mbr_buf))
-	    {
-	      errnum = ERR_BAD_PART_TABLE;
-	      return 0;
-	    }
-
-	  ext = -1;
-
-	  /*
-	   *  Scan table for partitions
-	   */
-	  for (i = 0; i < PC_SLICE_MAX; i++)
-	    {
-	      current_partition = ((slice_no << 16)
-				   | (current_partition & 0xFFFF));
-	      current_slice = PC_SLICE_TYPE (mbr_buf, i);
-	      part_start = part_offset + PC_SLICE_START (mbr_buf, i);
-	      part_length = PC_SLICE_LENGTH (mbr_buf, i);
 #ifndef STAGE1_5
-	      memmove (cur_part_desc, mbr_buf + PC_SLICE_OFFSET + (i << 4), 16);
-#endif
+    loop_start:
+      
+      cur_part_offset = part_offset;
+      cur_part_addr = BOOT_PART_TABLE + (entry << 4);
+#endif /* ! STAGE1_5 */
 
-	      /*
-	       *  Is this PC partition entry valid?
-	       */
-	      if (current_slice)
-		{
-		  /*
-		   *  Is this an extended partition?
-		   */
-		  if (IS_PC_SLICE_TYPE_EXTENDED (current_slice))
-		    {
-		      if (ext == -1)
-			{
-			  ext = i;
-			}
-		    }
-# ifndef STAGE1_5
-		  /*
-		   *  Display partition information
-		   */
-		  else if (flags)
-		    {
-		      current_partition |= 0xFFFF;
-		      if (! do_completion)
-			{
-			  printf ("   Partition num: %d, ", slice_no);
-		      
-			  if (! IS_PC_SLICE_TYPE_BSD (current_slice))
-			    check_and_print_mount ();
-			  else
-			    check_BSD_parts (1);
-			}
-		      else
-			{
-			  if (! IS_PC_SLICE_TYPE_BSD (current_slice))
-			    {
-			      char str[8];
-
-			      grub_sprintf (str, "%d)", slice_no);
-			      print_a_completion (str);
-			    }
-			  else
-			    check_BSD_parts (1);
-			}
-		      
-		      errnum = ERR_NONE;
-		    }
-# endif /* ! STAGE1_5 */
-		  /*
-		   *  If we've found the right partition, we're done
-		   */
-		  else if (part_no == slice_no
-			   || (part_no == 0xFF
-			       && IS_PC_SLICE_TYPE_BSD (current_slice)))
-		    {
-		      if ((current_partition & 0xFF00) != 0xFF00)
-			{
-			  if (IS_PC_SLICE_TYPE_BSD (current_slice))
-			    check_BSD_parts (0);
-			  else
-			    errnum = ERR_NO_PART;
-			}
-
-		      ext = -2;
-		      break;
-		    }
-		}
-
-	      /*
-	       *  If we're beyond the end of the standard PC partition
-	       *  range, change the numbering from one per table entry
-	       *  to one per valid entry.
-	       */
-	      if (slice_no < PC_SLICE_MAX
-		  || (! IS_PC_SLICE_TYPE_EXTENDED (current_slice)
-		      && current_slice != PC_SLICE_TYPE_NONE))
-		slice_no++;
-	    }
-
-	  part_offset = ext_offset + PC_SLICE_START (mbr_buf, ext);
-	  if (!ext_offset)
-	    ext_offset = part_offset;
-	}
-    }
-  else
-    {
-      /*
-       *  We're looking at a floppy disk
-       */
-      ext = -1;
-      if ((flags || (current_partition & 0xFF00) != 0xFF00)
-	  && check_BSD_parts (flags))
-	ext = -2;
-      else
+      /* If this is a valid partition...  */
+      if (current_slice)
 	{
-	  errnum = 0;
-	  if (!flags)
+#ifndef STAGE1_5
+	  /* Display partition information.  */
+	  if (flags && ! IS_PC_SLICE_TYPE_EXTENDED (current_slice))
 	    {
-	      if (current_partition == 0xFFFFFF
-		  || current_partition == 0xFF00FF)
+	      if (! do_completion)
 		{
-		  current_partition = 0xFFFFFF;
-		  ext = -2;
+		  if (current_drive & 0x80)
+		    grub_printf ("   Partition num: %d, ",
+				 current_partition >> 16);
+
+		  if (! IS_PC_SLICE_TYPE_BSD (current_slice))
+		    check_and_print_mount ();
+		  else
+		    {
+		      int got_part = 0;
+		      int saved_slice = current_slice;
+		      
+		      while (next ())
+			{
+			  if (bsd_part == 0xFF)
+			    break;
+			  
+			  if (! got_part)
+			    {
+			      grub_printf ("[BSD sub-partitions immediately follow]\n");
+			      got_part = 1;
+			    }
+			  
+			  grub_printf ("     BSD Partition num: \'%c\', ",
+				       bsd_part + 'a');
+			  check_and_print_mount ();
+			}
+
+		      if (! got_part)
+			grub_printf (" No BSD sub-partition found, partition type 0x%x\n",
+				     saved_slice);
+		      
+		      if (errnum)
+			{
+			  errnum = ERR_NONE;
+			  break;
+			}
+		      
+		      goto loop_start;
+		    }
+		}
+	      else
+		{
+		  if (bsd_part != 0xFF)
+		    {
+		      char str[16];
+		      
+		      if (! (current_drive & 0x80)
+			  || (dest_partition >> 16) == pc_slice)
+			grub_sprintf (str, "%c)", bsd_part + 'a');
+		      else
+			grub_sprintf (str, "%d,%c)",
+				      pc_slice, bsd_part + 'a');
+		      print_a_completion (str);
+		    }
+		  else if (! IS_PC_SLICE_TYPE_BSD (current_slice))
+		    {
+		      char str[8];
+		      
+		      grub_sprintf (str, "%d)", pc_slice);
+		      print_a_completion (str);
+		    }
 		}
 	    }
-# ifndef STAGE1_5
-	  else
-	    {
-	      current_partition = 0xFFFFFF;
-	      check_and_print_mount ();
-	      errnum = 0;
-	    }
-# endif /* STAGE1_5 */
+	  
+	  errnum = ERR_NONE;
+#endif /* ! STAGE1_5 */
+	  
+	  /* Check if this is the destination partition.  */
+	  if (! flags
+	      && (dest_partition == current_partition
+		  || ((dest_partition >> 16) == 0xFF
+		      && ((dest_partition >> 8) & 0xFF) == bsd_part)))
+	    return 1;
 	}
     }
 
-  if (!flags && (ext != -2) && (errnum == ERR_NONE))
-    errnum = ERR_NO_PART;
-
-  if (errnum != ERR_NONE)
-    return 0;
-
-  return 1;
+#ifndef STAGE1_5
+  if (flags)
+    {
+      if (! (current_drive & 0x80))
+	{
+	  current_partition = 0xFFFFFF;
+	  check_and_print_mount ();
+	}
+      
+      errnum = ERR_NONE;
+      return 1;
+    }
+#endif /* ! STAGE1_5 */
+  
+  return 0;
 }
 
 
@@ -683,16 +803,14 @@ set_device (char *device)
   int drive = (dev >> 24) & 0xFF;
   int partition = dev & 0xFFFFFF;
 
+  /* If DRIVE is disabled (0xFF), use SAVED_DRIVE instead.  */
   if (drive == 0xFF)
-    {
-      current_drive = saved_drive;
-      current_partition = saved_partition;
-    }
+    current_drive = saved_drive;
   else
-    {
-      current_drive = drive;
-      current_partition = partition;
-    }
+    current_drive = drive;
+
+  /* The `partition' part must always have a valid number.  */
+  current_partition = partition;
   
   return device + sizeof (unsigned long);
   
@@ -736,7 +854,7 @@ set_device (char *device)
 	    errnum = ERR_NUMBER_PARSING;
 
 	  if (ch == 'n')
-	    current_drive = 0x20;
+	    current_drive = NETWORK_DRIVE;
 	  else
 	    {
 	      safe_parse_maxint (&device, (int *) &current_drive);
@@ -777,10 +895,11 @@ set_device (char *device)
 
 	      current_partition = (current_partition << 16) + 0xFFFF;
 
-	      if (*device == ','
-		  && *(device + 1) >= 'a' && *(device + 1) <= 'h')
+	      if (*device == ',')
+		device++;
+	      
+	      if (*device >= 'a' && *device <= 'h')
 		{
-		  device++;
 		  current_partition = (((*(device++) - 'a') << 8)
 				       | (current_partition & 0xFF00FF));
 		}
@@ -803,78 +922,6 @@ set_device (char *device)
 	    }
 	}
     }
-  else
-    {
-      char ch;
-
-      /* A Mach-style absolute partition name. */
-      ch = *device;
-      if ((*device != 'f' && *device != 'h') ||
-	  (device += 2, (*(device - 1) != 'd')))
-	errnum = ERR_DEV_FORMAT;
-      else
-	safe_parse_maxint (&device, (int *) &current_drive);
-
-      disk_choice = 0;
-      if (ch != 'f')
-	current_drive += 0x80;
-
-      if (errnum)
-	return 0;
-
-      if (*device == '/')
-	{
-	  part_choice = PART_CHOSEN;
-	  result = 1;
-	}
-      else if (*device == 's')
-	{
-	  /* An absolute PC partition. */
-	  disk_choice = 0;
-	  part_choice ++;
-	  device ++;
-
-	  if (*device >= '0' && *device <= '9')
-	    {
-	      part_choice ++;
-	      current_partition = 0;
-
-	      if (!(current_drive & 0x80) ||
-		  !safe_parse_maxint (&device, (int *) &current_partition) ||
-		  (--current_partition) > 254)
-		{
-		  errnum = ERR_DEV_FORMAT;
-		  return 0;
-		}
-
-	      current_partition = (current_partition << 16) + 0xFFFF;
-
-	      if (*device >= 'a' && *device <= 'h')
-		{
-		  /* A BSD partition within the slice. */
-		  current_partition = (((*(device ++) - 'a') << 8)
-				       | (current_partition & 0xFF00FF));
-		}
-	    }
-	}
-      else if (*device >= 'a' && *device <= 'h')
-	{
-	  /* An absolute BSD partition. */
-	  part_choice ++;
-	  current_partition = ((*(device ++) - 'a') << 8) | 0xFF00FF;
-	}
-
-      if (*device == '/')
-	{
-	  if (part_choice == PART_DISK)
-	    {
-	      current_partition = saved_partition;
-	      part_choice ++;
-	    }
-
-	  result = 1;
-	}
-    }
 
   if (! sane_partition ())
     return 0;
@@ -917,11 +964,10 @@ set_bootdev (int hdbias)
 {
   int i, j;
 
-  /*
-   *  Set chainloader boot device.
-   */
-  memmove ((char *) (BOOTSEC_LOCATION - 16), cur_part_desc, 16);
-
+  /* Save the boot partition for chain-loading.  */
+  boot_part_offset = cur_part_offset;
+  boot_part_addr = cur_part_addr;
+  
   /*
    *  Set BSD boot device.
    */
@@ -963,11 +1009,6 @@ setup_part (char *filename)
   
 #else /* ! STAGE1_5 */
   
-  /* FIXME: decide on syntax for blocklist vs. old-style vs. /dev/hd0s1 */
-  /* Strip any leading /dev. */
-  if (substring ("/dev/", filename) < 1)
-    filename += 5;
-
   if (*filename == '(')
     {
       if ((filename = set_device (filename)) == 0)
@@ -1035,7 +1076,7 @@ print_fsys_type (void)
       if (current_partition == 0xFFFFFF)
 	printf ("using whole disk\n");
       else
-	printf ("partition type 0x%x\n", current_slice);
+	printf ("partition type 0x%x\n", current_slice & 0xFF);
     }
 }
 #endif /* STAGE1_5 */
@@ -1124,15 +1165,19 @@ print_completions (int is_filename, int is_completion)
       if (! is_completion)
 	grub_putchar ('\n');
       
+      print_error ();
       do_completion = 0;
-      return unique - 1;
+      if (errnum)
+	return -1;
+      else
+	return unique - 1;
     }
 
   if (*buf == '/' || (ptr = set_device (buf)) || incomplete)
     {
       errnum = 0;
 
-      if (*buf != '/' && (incomplete || ! *ptr))
+      if (*buf == '(' && (incomplete || ! *ptr))
 	{
 	  if (! part_choice)
 	    {
@@ -1190,7 +1235,19 @@ print_completions (int is_filename, int is_completion)
 	  else
 	    {
 	      /* partition completions */
-	      if (part_choice == PART_DISK)
+	      if (part_choice == PART_CHOSEN
+		  && open_partition ()
+		  && ! IS_PC_SLICE_TYPE_BSD (current_slice))
+		{
+		  unique = 1;
+		  ptr = buf + grub_strlen (buf);
+		  if (*(ptr - 1) != ')')
+		    {
+		      *ptr++ = ')';
+		      *ptr = 0;
+		    }
+		}
+	      else
 		{
 		  if (! is_completion)
 		    grub_printf (" Possible partitions are:\n");
@@ -1204,22 +1261,9 @@ print_completions (int is_filename, int is_completion)
 		      grub_strcpy (ptr, unique_string);
 		    }
 		}
-	      else
-		{
-		  if (open_partition ())
-		    {
-		      unique = 1;
-		      ptr = buf + grub_strlen (buf);
-		      if (*(ptr - 1) != ')')
-			{
-			  *ptr++ = ')';
-			  *ptr = 0;
-			}
-		    }
-		}
 	    }
 	}
-      else if (*ptr == '/')
+      else if (ptr && *ptr == '/')
 	{
 	  /* filename completions */
 	  if (! is_completion)
@@ -1268,7 +1312,10 @@ print_completions (int is_filename, int is_completion)
 
   print_error ();
   do_completion = 0;
-  return unique - 1;
+  if (errnum)
+    return -1;
+  else
+    return unique - 1;
 }
 #endif /* STAGE1_5 */
 
@@ -1391,11 +1438,11 @@ int
 grub_read (char *buf, int len)
 {
   /* Make sure "filepos" is a sane value */
-  if ((filepos < 0) | (filepos > filemax))
+  if ((filepos < 0) || (filepos > filemax))
     filepos = filemax;
 
   /* Make sure "len" is a sane value */
-  if ((len < 0) | (len > (filemax - filepos)))
+  if ((len < 0) || (len > (filemax - filepos)))
     len = filemax - filepos;
 
   /* if target file position is past the end of
@@ -1486,8 +1533,18 @@ grub_read (char *buf, int len)
   return (*(fsys_table[fsys_type].read_func)) (buf, len);
 }
 
-
 #ifndef STAGE1_5
+/* Reposition a file offset.  */
+int
+grub_seek (int offset)
+{
+  if (offset > filemax || offset < 0)
+    return -1;
+
+  filepos = offset;
+  return offset;
+}
+
 int
 dir (char *dirname)
 {
@@ -1513,3 +1570,15 @@ dir (char *dirname)
   return (*(fsys_table[fsys_type].dir_func)) (dirname);
 }
 #endif /* STAGE1_5 */
+
+void 
+grub_close (void)
+{
+#ifndef NO_BLOCK_FILES
+  if (block_file)
+    return;
+#endif /* NO_BLOCK_FILES */
+  
+  if (fsys_table[fsys_type].close_func != 0)
+    (*(fsys_table[fsys_type].close_func)) ();
+}
