@@ -2,7 +2,7 @@
 /*
  *  GRUB  --  GRand Unified Bootloader
  *  Copyright (C) 1996  Erich Boleyn  <erich@uruk.org>
- *  Copyright (C) 1999  Free Software Foundation, Inc.
+ *  Copyright (C) 1999, 2000 Free Software Foundation, Inc.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,9 @@
  *  Generic defines to use anywhere
  */
 
+#ifndef GRUB_SHARED_HEADER
+#define GRUB_SHARED_HEADER	1
+
 #include <config.h>
 
 /* Add an underscore to a C symbol in assembler code if needed. */
@@ -66,6 +69,8 @@ extern char *grub_scratch_mem;
 #define BUFFERADDR  RAW_ADDR (0x70000)
 #define BUFFERSEG   RAW_SEG (0x7000)
 
+#define BOOT_PART_TABLE	RAW_ADDR (0x07be)
+
 /*
  *  BIOS disk defines
  */
@@ -117,46 +122,54 @@ extern char *grub_scratch_mem;
 #define MENU_BUF		(UNIQUE_BUF + UNIQUE_BUFLEN)
 #define MENU_BUFLEN		(0x8000 + PASSWORD_BUF - UNIQUE_BUF)
 
+/* The size of the drive map.  */
+#define DRIVE_MAP_SIZE		8
+
+/* The size of the key map.  */
+#define KEY_MAP_SIZE		128
+
+/* The size of the io map.  */
+#define IO_MAP_SIZE		128
+
 /*
  *  Linux setup parameters
  */
 
+#define LINUX_MAGIC_SIGNATURE		0x53726448	/* "HdrS" */
+#define LINUX_DEFAULT_SETUP_SECTS	4
+#define LINUX_FLAG_CAN_USE_HEAP		0x80
+#define LINUX_INITRD_MAX_ADDRESS	0x3C000000
+#define LINUX_MAX_SETUP_SECTS		63
+#define LINUX_BOOT_LOADER_TYPE		0x71
+#define LINUX_HEAP_END_OFFSET		(0x7F00 - 0x200)
+
 #define LINUX_STAGING_AREA        RAW_ADDR (0x100000)
 #define LINUX_SETUP               RAW_ADDR (0x90000)
-#define LINUX_SETUP_MAXLEN        0x1E00
 #define LINUX_KERNEL              RAW_ADDR (0x10000)
 #define LINUX_KERNEL_MAXLEN       0x7F000
 #define LINUX_SETUP_SEG           0x9020
 #define LINUX_INIT_SEG            0x9000
-#define LINUX_KERNEL_LEN_OFFSET   0x1F4
-#define LINUX_SETUP_LEN_OFFSET    0x1F1
-#define LINUX_SETUP_STACK         0x3FF4
+#define LINUX_SETUP_STACK         0x7F00
 
-#define LINUX_SETUP_LOADER        0x210
-#define LINUX_SETUP_LOAD_FLAGS    0x211
 #define LINUX_FLAG_BIG_KERNEL     0x1
-#define LINUX_SETUP_CODE_START    0x214
-#define LINUX_SETUP_INITRD        0x218
 
 /* Linux's video mode selection support. Actually I hate it!  */
-#define LINUX_VID_MODE_OFFSET	0x1FA
 #define LINUX_VID_MODE_NORMAL	0xFFFF
 #define LINUX_VID_MODE_EXTENDED	0xFFFE
 #define LINUX_VID_MODE_ASK	0xFFFD
 
-#define CL_MY_LOCATION  RAW_ADDR (0x92000)
-#define CL_MY_END_ADDR  RAW_ADDR (0x920FF)
-#define CL_MAGIC_ADDR   RAW_ADDR (0x90020)
+#define CL_MY_LOCATION  RAW_ADDR (0x97F00)
+#define CL_MY_END_ADDR  RAW_ADDR (0x97FFF)
 #define CL_MAGIC        0xA33F
 #define CL_BASE_ADDR    RAW_ADDR (0x90000)
-#define CL_OFFSET       RAW_ADDR (0x90022)
 
 /*
  *  General disk stuff
  */
 
-#define SECTOR_SIZE          0x200
-#define BIOS_FLAG_FIXED_DISK 0x80
+#define SECTOR_SIZE		0x200
+#define SECTOR_BITS		9
+#define BIOS_FLAG_FIXED_DISK	0x80
 
 #define BOOTSEC_LOCATION     RAW_ADDR (0x7C00)
 #define BOOTSEC_SIGNATURE    0xAA55
@@ -165,6 +178,10 @@ extern char *grub_scratch_mem;
 #define BOOTSEC_PART_OFFSET  0x1BE
 #define BOOTSEC_PART_LENGTH  0x40
 #define BOOTSEC_SIG_OFFSET   0x1FE
+#define BOOTSEC_LISTSIZE     8
+
+/* Not bad, perhaps.  */
+#define NETWORK_DRIVE	0x20
 
 /*
  *  GRUB specific information
@@ -173,10 +190,12 @@ extern char *grub_scratch_mem;
 
 #include <stage1.h>
 
-#define STAGE2_VER_MAJ_OFFS  0x6
-#define STAGE2_INSTALLPART   0x8
-#define STAGE2_STAGE2_ID     0xc
-#define STAGE2_VER_STR_OFFS  0xd
+#define STAGE2_VER_MAJ_OFFS	0x6
+#define STAGE2_INSTALLPART	0x8
+#define STAGE2_SAVED_ENTRYNO	0xc
+#define STAGE2_STAGE2_ID	0x10
+#define STAGE2_FORCE_LBA	0x11
+#define STAGE2_VER_STR_OFFS	0x12
 
 /* Stage 2 identifiers */
 #define STAGE2_ID_STAGE2		0
@@ -184,6 +203,7 @@ extern char *grub_scratch_mem;
 #define STAGE2_ID_E2FS_STAGE1_5		2
 #define STAGE2_ID_FAT_STAGE1_5		3
 #define STAGE2_ID_MINIX_STAGE1_5	4
+#define STAGE2_ID_REISERFS_STAGE1_5	5
 
 #ifndef STAGE1_5
 # define STAGE2_ID	STAGE2_ID_STAGE2
@@ -196,6 +216,8 @@ extern char *grub_scratch_mem;
 #  define STAGE2_ID	STAGE2_ID_FAT_STAGE1_5
 # elif defined(FSYS_MINIX)
 #  define STAGE2_ID	STAGE2_ID_MINIX_STAGE1_5
+# elif defined(FSYS_REISERFS)
+#  define STAGE2_ID	STAGE2_ID_REISERFS_STAGE1_5
 # else
 #  error "unknown Stage 2"
 # endif
@@ -211,8 +233,8 @@ extern char *grub_scratch_mem;
 #define PROT_MODE_DSEG  0x10
 #define PSEUDO_RM_CSEG	0x18
 #define PSEUDO_RM_DSEG	0x20
-#define STACKOFF	0x2000 - 0x10
-#define PROTSTACKINIT   FSYS_BUF - 0x10
+#define STACKOFF	(0x2000 - 0x10)
+#define PROTSTACKINIT   (FSYS_BUF - 0x10)
 
 
 /*
@@ -277,17 +299,54 @@ extern char *grub_scratch_mem;
 # endif /* ! A_STANDOUT */
 #endif /* ! A_REVERSE */
 
+/* Define ACS_* ourselves, since the definitions are not consistent among
+   various curses implementations.  */
+#undef ACS_ULCORNER
+#undef ACS_URCORNER
+#undef ACS_LLCORNER
+#undef ACS_LRCORNER
+#undef ACS_HLINE
+#undef ACS_VLINE
+#undef ACS_LARROW
+#undef ACS_RARROW
+#undef ACS_UARROW
+#undef ACS_DARROW
+
+#define ACS_ULCORNER	'+'
+#define ACS_URCORNER	'+'
+#define ACS_LLCORNER	'+'
+#define ACS_LRCORNER	'+'
+#define ACS_HLINE	'-'
+#define ACS_VLINE	'|'
+#define ACS_LARROW	'<'
+#define ACS_RARROW	'>'
+#define ACS_UARROW	'^'
+#define ACS_DARROW	'v'
+
 /* Special graphics characters for IBM displays. */
-#define DISP_UL         218
-#define DISP_UR         191
-#define DISP_LL         192
-#define DISP_LR         217
-#define DISP_HORIZ      196
-#define DISP_VERT       179
-#define DISP_LEFT       0x1b
-#define DISP_RIGHT      0x1a
-#define DISP_UP         0x18
-#define DISP_DOWN       0x19
+#ifdef GRUB_UTIL
+# define DISP_UL	ACS_ULCORNER
+# define DISP_UR	ACS_URCORNER
+# define DISP_LL	ACS_LLCORNER
+# define DISP_LR	ACS_LRCORNER
+# define DISP_HORIZ	ACS_HLINE
+# define DISP_VERT	ACS_VLINE
+# define DISP_LEFT	ACS_LARROW
+# define DISP_RIGHT	ACS_RARROW
+# define DISP_UP	ACS_UARROW
+# define DISP_DOWN	ACS_DARROW
+#else /* ! GRUB_UTIL */
+# define DISP_UL	218
+# define DISP_UR	191
+# define DISP_LL	192
+# define DISP_LR	217
+# define DISP_HORIZ	196
+# define DISP_VERT	179
+# define DISP_LEFT	0x1b
+# define DISP_RIGHT	0x1a
+# define DISP_UP	0x18
+# define DISP_DOWN	0x19
+#endif /* ! GRUB_UTIL */
 
 /* Remap some libc-API-compatible function names so that we prevent
    circularararity. */
@@ -320,6 +379,38 @@ extern char *grub_scratch_mem;
 #include "mb_header.h"
 #include "mb_info.h"
 
+/* For the Linux/i386 boot protocol version 2.02.  */
+struct linux_kernel_header
+{
+  char code1[0x0020];
+  unsigned short cl_magic;		/* Magic number 0xA33F */
+  unsigned short cl_offset;		/* The offset of command line */
+  char code2[0x01F1 - 0x0020 - 2 - 2];
+  unsigned char setup_sects;		/* The size of the setup in sectors */
+  unsigned short root_flags;		/* If the root is mounted readonly */
+  unsigned short syssize;		/* obsolete */
+  unsigned short swap_dev;		/* obsolete */
+  unsigned short ram_size;		/* obsolete */
+  unsigned short vid_mode;		/* Video mode control */
+  unsigned short root_dev;		/* Default root device number */
+  unsigned short boot_flag;		/* 0xAA55 magic number */
+  unsigned short jump;			/* Jump instruction */
+  unsigned long header;			/* Magic signature "HdrS" */
+  unsigned short version;		/* Boot protocol version supported */
+  unsigned long realmode_swtch;		/* Boot loader hook */
+  unsigned long start_sys;		/* Points to kernel version string */
+  unsigned char type_of_loader;		/* Boot loader identifier */
+  unsigned char loadflags;		/* Boot protocol option flags */
+  unsigned short setup_move_size;	/* Move to high memory size */
+  unsigned long code32_start;		/* Boot loader hook */
+  unsigned long ramdisk_image;		/* initrd load address */
+  unsigned long ramdisk_size;		/* initrd size */
+  unsigned long bootsect_kludge;	/* obsolete */
+  unsigned short heap_end_ptr;		/* Free memory after setup end */
+  unsigned short pad1;			/* Unused */
+  unsigned long cmd_line_ptr;		/* Points to the kernel command line */
+} __attribute__ ((packed));
+
 /* Memory map address range descriptor used by GET_MMAP_ENTRY. */
 struct mmar_desc
 {
@@ -365,14 +456,23 @@ typedef enum
   ERR_UNRECOGNIZED,
   ERR_WONT_FIT,
   ERR_WRITE,
+  ERR_BAD_ARGUMENT,
+  ERR_UNALIGNED,
+  ERR_PRIVILEGED,
 
   MAX_ERR_NUM
 } grub_error_t;
 
 extern unsigned long install_partition;
 extern unsigned long boot_drive;
+extern unsigned long install_second_sector;
+extern unsigned long boot_part_addr;
+extern unsigned long boot_part_offset;
+extern int saved_entryno;
+extern unsigned char force_lba;
 extern char version_string[];
 extern char config_file[];
+extern unsigned long linux_text_len;
 
 #ifdef GRUB_UTIL
 /* If not using config file, this variable is set to zero,
@@ -384,17 +484,14 @@ extern int use_curses;
 extern int verbose;
 /* The flag for read-only.  */
 extern int read_only;
-/* If this flag is true, assume that there is no floppy.  */
-extern int no_floppy;
-/* If this flag is true, probe the second floppy drive.  */
-extern int probe_second_floppy;
+/* The number of floppies to be probed.  */
+extern int floppy_disks;
 /* The map between BIOS drives and UNIX device file names.  */
 extern char **device_map;
+/* The filename which stores the information about a device map.  */
+extern char *device_map_file;
 /* The array of geometries.  */
 extern struct geometry *disks;
-/* Check if DEVICE can be read. If an error occurs, return zero,
-   otherwise return non-zero.  */
-extern int check_device (const char *device);
 /* Assign DRIVE to a device name DEVICE.  */
 extern void assign_device_name (int drive, const char *device);
 #endif
@@ -403,7 +500,9 @@ extern void assign_device_name (int drive, const char *device);
 /* GUI interface variables. */
 extern int fallback_entry;
 extern int default_entry;
+extern int current_entryno;
 extern char *password;
+extern int auth;
 extern char commands[];
 #endif
 
@@ -414,8 +513,8 @@ extern int compressed_file;
 
 #ifndef STAGE1_5
 /* instrumentation variables */
-extern void (*disk_read_hook) (int);
-extern void (*disk_read_func) (int);
+extern void (*disk_read_hook) (int, int, int);
+extern void (*disk_read_func) (int, int, int);
 /* The flag for debug mode.  */
 extern int debug;
 /* Color settings */
@@ -427,10 +526,6 @@ extern unsigned long current_partition;
 
 extern int fsys_type;
 
-#ifndef NO_BLOCK_FILES
-extern int block_file;
-#endif /* NO_BLOCK_FILES */
-
 /* The information for a disk geometry. The CHS information is only for
    DOS/Partition table compatibility, and the real number of sectors is
    stored in TOTAL_SECTORS.  */
@@ -470,6 +565,7 @@ extern unsigned long saved_drive;
 extern unsigned long saved_partition;
 #ifndef STAGE1_5
 extern unsigned long saved_mem_upper;
+extern unsigned long extended_memory;
 #endif
 
 /*
@@ -491,10 +587,36 @@ void cmain (void);
 /* Halt the processor (called after an unrecoverable error). */
 void stop (void) __attribute__ ((noreturn));
 
+/* Reboot the system.  */
+void grub_reboot (void) __attribute__ ((noreturn));
+
+/* Halt the system, using APM if possible. If NO_APM is true, don't use
+   APM even if it is available.  */
+void grub_halt (int no_apm) __attribute__ ((noreturn));
+
+/* Copy MAP to the drive map and set up int13_handler.  */
+void set_int13_handler (unsigned short *map);
+
+/* Set up int15_handler.  */
+void set_int15_handler (void);
+
+/* Restore the original int15 handler.  */
+void unset_int15_handler (void);
+
+/* Track the int13 handler to probe I/O address space.  */
+void track_int13 (int drive);
+
+/* The key map.  */
+extern unsigned short bios_key_map[];
+extern unsigned short ascii_key_map[];
+extern unsigned short io_map[];
+
 /* calls for direct boot-loader chaining */
-void chain_stage1 (int segment, int offset, int part_table_addr)
+void chain_stage1 (unsigned long segment, unsigned long offset,
+		   unsigned long part_table_addr)
+     __attribute__ ((noreturn));
+void chain_stage2 (unsigned long segment, unsigned long offset)
      __attribute__ ((noreturn));
-void chain_stage2 (int segment, int offset) __attribute__ ((noreturn));
 
 /* do some funky stuff, then boot linux */
 void linux_boot (void) __attribute__ ((noreturn));
@@ -503,7 +625,7 @@ void linux_boot (void) __attribute__ ((noreturn));
 void big_linux_boot (void) __attribute__ ((noreturn));
 
 /* booting a multiboot executable */
-void multi_boot (int start, int mbi) __attribute__ ((noreturn));
+void multi_boot (int start, int mb_info) __attribute__ ((noreturn));
 
 /* If LINEAR is nonzero, then set the Intel processor to linear mode.
    Otherwise, bit 20 of all memory accesses is always forced to zero,
@@ -531,6 +653,9 @@ int currticks (void);
 /* Clear the screen. */
 void cls (void);
 
+/* The console part of cls.  */
+void console_cls (void);
+
 #ifndef GRUB_UTIL
 /* Turn off cursor. */
 void nocursor (void);
@@ -541,28 +666,43 @@ void nocursor (void);
    (RET & 0xff) is y. */
 int getxy (void);
 
+/* The console part of getxy.  */
+int console_getxy (void);
+
 /* Set the cursor position. */
 void gotoxy (int x, int y);
 
+/* The console part of gotoxy.  */
+void console_gotoxy (int x, int y);
+
 /* Displays an ASCII character.  IBM displays will translate some
    characters to special graphical ones (see the DISP_* constants). */
 void grub_putchar (int c);
 
+/* The console part of grub_putchar.  */
+void console_putchar (int c);
+
 /* Wait for a keypress, and return its packed BIOS/ASCII key code.
    Use ASCII_CHAR(ret) to extract the ASCII code. */
 int getkey (void);
 
+/* The console part of getkey.  */
+int console_getkey (void);
+
 /* Like GETKEY, but doesn't block, and returns -1 if no keystroke is
    available. */
 int checkkey (void);
 
+/* The console part of checkkey.  */
+int console_checkkey (void);
+
 /* Sets text mode character attribute at the cursor position.  See A_*
    constants defined above. */
 void set_attrib (int attr);
 
 /* Low-level disk I/O */
 int get_diskinfo (int drive, struct geometry *geometry);
-int biosdisk (int read, int drive, struct geometry *geometry,
+int biosdisk (int subfunc, int drive, struct geometry *geometry,
 	      int sector, int nsec, int segment);
 void stop_floppy (void);
 
@@ -573,6 +713,7 @@ void stop_floppy (void);
 #define BUILTIN_CMDLINE		0x1	/* Run in the command-line.  */
 #define BUILTIN_MENU		0x2	/* Run in the menu.  */
 #define BUILTIN_TITLE		0x4	/* Only for the command title.  */
+#define BUILTIN_SCRIPT		0x8	/* Run in the script.  */
 
 /* The table for a builtin.  */
 struct builtin
@@ -602,17 +743,30 @@ typedef enum
   KERNEL_TYPE_FREEBSD,		/* FreeBSD.  */
   KERNEL_TYPE_NETBSD,		/* NetBSD.  */
   KERNEL_TYPE_CHAINLOADER	/* Chainloader.  */
-} kernel_t;
+}
+kernel_t;
 
 extern kernel_t kernel_type;
+extern int show_menu;
 extern int grub_timeout;
 
+/* Control the auto fill mode.  */
+extern int auto_fill;
+
+/* This variable specifies which console should be used.  */
+extern int terminal;
+
+#define TERMINAL_CONSOLE	(1 << 0)	/* keyboard and screen */
+#define TERMINAL_SERIAL		(1 << 1)	/* serial console */
+
+#define TERMINAL_DUMB		(1 << 16)	/* dumb terminal */
+
 void init_builtins (void);
 void init_config (void);
 char *skip_to (int after_equal, char *cmdline);
 struct builtin *find_command (char *command);
-void print_cmdline_message (void);
-void enter_cmdline (char *heap);
+void print_cmdline_message (int forever);
+void enter_cmdline (char *heap, int forever);
 int run_script (char *script, char *heap);
 #endif
 
@@ -622,14 +776,36 @@ int grub_sprintf (char *buffer, const char *format, ...);
 int grub_tolower (int c);
 int grub_isspace (int c);
 int grub_strncat (char *s1, const char *s2, int n);
-char *grub_memmove (char *to, const char *from, int len);
+void *grub_memmove (void *to, const void *from, int len);
 void *grub_memset (void *start, int c, int len);
+int grub_strncat (char *s1, const char *s2, int n);
 char *grub_strstr (const char *s1, const char *s2);
 int grub_memcmp (const char *s1, const char *s2, int n);
 int grub_strcmp (const char *s1, const char *s2);
 int grub_strlen (const char *str);
 char *grub_strcpy (char *dest, const char *src);
 
+#ifndef GRUB_UTIL
+typedef unsigned long grub_jmp_buf[6];
+#else
+/* In the grub shell, use the libc jmp_buf instead.  */
+# include <setjmp.h>
+# define grub_jmp_buf jmp_buf
+#endif
+
+#ifdef GRUB_UTIL
+# define grub_setjmp	setjmp
+# define grub_longjmp	longjmp
+#else /* ! GRUB_UTIL */
+int grub_setjmp (grub_jmp_buf env);
+void grub_longjmp (grub_jmp_buf env, int val);
+#endif /* ! GRUB_UTIL */
+
+/* The environment for restarting Stage 2.  */
+extern grub_jmp_buf restart_env;
+/* The environment for restarting the command-line interface.  */
+extern grub_jmp_buf restart_cmdline_env;
+
 /* misc */
 void init_page (void);
 void print_error (void);
@@ -637,9 +813,11 @@ char *convert_to_ascii (char *buf, int c, ...);
 int get_cmdline (char *prompt, char *cmdline, int maxlen,
 		 int echo_char, int history);
 int substring (char *s1, char *s2);
+int nul_terminate (char *str);
 int get_based_digit (int c, int base);
 int safe_parse_maxint (char **str_ptr, int *myint_ptr);
 int memcheck (int start, int len);
+int translate_keycode (int c);
 
 #ifndef NO_DECOMPRESSION
 /* Compression support. */
@@ -655,6 +833,11 @@ char *set_device (char *device);
 int open_device (void);
 int real_open_partition (int flags);
 int open_partition (void);
+int next_partition (unsigned long drive, unsigned long dest,
+		    unsigned long *partition, int *type,
+		    unsigned long *start, unsigned long *len,
+		    unsigned long *offset, int *entry,
+		    unsigned long *ext_offset, char *buf);
 
 /* Sets device to the one represented by the SAVED_* parameters. */
 int make_saved_active (void);
@@ -667,9 +850,15 @@ int set_partition_hidden_flag (int hidden);
 int grub_open (char *filename);
 
 /* Read LEN bytes into BUF from the file that was opened with
-   GRUB_OPEN.  If LEN is -1, read all the remaining data in the file */
+   GRUB_OPEN.  If LEN is -1, read all the remaining data in the file.  */
 int grub_read (char *buf, int len);
 
+/* Reposition a file offset.  */
+int grub_seek (int offset);
+
+/* Close a file.  */
+void grub_close (void);
+
 /* List the contents of the directory that was opened with GRUB_OPEN,
    printing all completions. */
 int dir (char *dirname);
@@ -689,7 +878,14 @@ void copy_current_part_entry (char *buf);
 #ifndef STAGE1_5
 void bsd_boot (kernel_t type, int bootdev, char *arg)
      __attribute__ ((noreturn));
-kernel_t load_image (char *kernel, char *arg);
+
+/* Define flags for load_image here.  */
+/* Don't pass a Linux's mem option automatically.  */
+#define KERNEL_LOAD_NO_MEM_OPTION	(1 << 0)
+
+kernel_t load_image (char *kernel, char *arg, kernel_t suggested_type,
+		     unsigned long load_flags);
+
 int load_module (char *module, char *arg);
 int load_initrd (char *initrd);
 #endif
@@ -697,3 +893,5 @@ int load_initrd (char *initrd);
 void init_bios_info (void);
 
 #endif /* ASM_FILE */
+
+#endif /* ! GRUB_SHARED_HEADER */
